/*
 * 	br2 -- Bias reduction squared
 *	version 1.06a (September 25, 2009)
 *
 *  LICENSE: Released under GNU General Public License, v2 (see COPYING.txt)
 *
 *  See file README for explanation of command line options
 *  See file NOTES for recent changes
 *
 *
 *  File:   br2.cpp
 *  Author: Apostolos Dimitromanolakis
 *
 *  Created on October 17, 2008, 5:22 PM
 */

#include "br2.h"
#include "plinkModule.h"
#include "random_gen.h"
#include "options.h"
#include "cluster.h"
#include "dummy.h"
#include "plink-code/plink.h"

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;

const char version_string[500] =
    "   br2 - Bias reduction squared\n"
    "   version 1.06a (September 25, 2009)\n";

namespace br2options {

    // alpha cutoff value
    double alpha = 1e-6;

    // if >0 then only top <k> SNPs are analyzed
    int maxSNPs = -1;

    // Type of significance test
    char testTypes[8][40] = {
	    "minTRENDGENO", "TREND", "GENO", "LOGISTIC", "ALLELIC", "DUMMY", "RPLUGIN", "LINEAR"
    };

    enum TEST_TYPES
    {
	minTRENDGENO = 0,
	TREND = 1,
	GENO = 2,
	LOGISTIC = 3,
	ALLELIC = 4,
	RPLUGIN = 5,
	LINEAR = 6,
	DUMMY = 66
    };

    // Default significance test type is TREND
    int testType = TREND;

    enum MODEL_TYPES
    {
	LOGISTIC_MODEL = 0,
	GENOTYPIC_MODEL = 1,
	ALLELIC_MODEL = 2,
	RPLUGIN_MODEL = 3,
	DUMMY_MODEL = 66
    };

    // Select between LOGISTIC and GENOTYPIC model.
    int estimationType = LOGISTIC_MODEL;

    // QTL is set to 1 if we are running analysis with QTL.
    int QuantTrait = 0;

    // Number of bootstraps to perform at each level
    int Level0Bootstraps = 100;
    int Level1Bootstraps = 0;
    int Level2Bootstraps = 100;

    // Directory for temporary files (must end with /)
    string tmpdir = "";

    // File to store output results
    string resultsFile = "out.br2";

    // Run more silently?
    int silent = 1;

    // If 1 then don't do truncate UBOOT to 0
    int notruncate = 0;

    // Number of cpus to use on this machine , 0 = single threaded
    int numCpu = 0;

    // Filter SNPs based on minor allele frequency , 0 means no filtering
    double mafFilter = 0;
    double mafFilterCases = 0;
    double mafFilterControls = 0;

    // File containing SNPS to exclude from analysis
    string excludefile;

    // File containing a list of SNPs to keep
    string extractfile;

    // Keep only top k SNPs to speedup bootstrap (if -1 then keep all)
    int keeptop = -1;

    // Keep the significant and a random selection of snps (for testing only)
    int keeprandom = -1;

    // Use plink internally,
    // Defaults to true , since code has been verified
    int usePlinkCode = 1;

    // Output a file containing statistics (maf / mafCases / mafControls) on all SNPs
    int printStatsFile = 0;

    // Number of cases that we found an abnormally large OR.
    int flagBadOR = 0;

    // Limit for Warning of Bad Log OR
    double WarnLogORLimit = 3;

    // Set if test-run with internally random data
    int dummyRun = 0;

    // Debugging: output a list of all bias values
    int logAllBias = 0;

    // Whether we are using covariates
    bool haveCovariates = false;

    // The name of the R script plugin, if we are using an R plugin
    string RScriptFile = "";

    // Debug flag
    int debug = 0;

    // Whether to report odds ratios or log odd ratios.
    int reportLogOR = 0;

    // Whether to apply MAF-specific OR correction
    int mafCorrection = 1;

    // Whether to use internal regression code
    int useInternalRegression = 1;


    // Clustering options

    // How many instances to split load to
    int clusterN = 0;

    // Which instance is the one we are running
    int clusterK = 0;

    // Whether to join the results of a previous cluster run
    int clusterJoin = 0;

    // Cluster results file
    ofstream clusterFile;
};

using namespace br2options;


// Parameters related to plink and covariate files
namespace par {
    bool clist_selection = false;
    bool clist_selection_name = false;
    bool clist_selection_number = false;

    string covar_filename;
    string clist_selection_string = "";

    string covariate_command_line;

    string covariate_header;
    vector<string> covariate_names;

    // Missing phenotype value (can be any string)
    string missing_phenotype = "-9";

    string alternative_phenotype_file = "";
};

// NCBI Class for random number generation
CRandom RNG;

// Locus contains the list of SNPs
int nlocus;
vector<Locus*> locus;

// The individuals along with genotype data
int npersons;
vector<Individual*> persons;

// Object used to run plink and parse output files
plinkModule plinkRunner;

// Internal code of plink for full-model association
Plink plinkInternalCode;

/////////////////////////////////////////////////////

// P-values and OR for significant SNPs (naive estimate).

// Number of significant SNPs in Original dataset
int Sn;

// Un: Naives estimates Un[i].odds = log odds ratio
vector<Results> Un;

// Total number of L1/L2 samples computed
int totalSamples = 0;

/////////////////////////////////////////////////////

/*
 * computeBias: computes the bias from a set of bootstrap samples
 *
 * Input: Un(naive estimate), L0: Bootstrap samples, level: which level of the bootstrap is this
 * Returns: Uboot : bias reduced estimates.
 *
 */


vector<double> computeBias(const vector<Results> &Un, const vector<vector< Results > > &L0, int level)
{
    vector<double> biasAll;
    char s[1024];

    int Sn = Un.size();
    biasAll.reserve(Sn);

    int N = L0.size();

    if(debug) cout << "Computing uboot using " << N << " iterations,,,\n";

    if(debug) { for (int k = 0; k < Sn; k++) { cout << Un[k].odds << " "; } cout << "...\n"; }

    for (int k = 0; k < Sn; k++) { // For each significant snp
	vector<double> bias, Ud, Ue;

	for (int b = 0; b < N; b++)
	    if (L0[b].size() > k && L0[b][k].UdOR != NA && L0[b][k].UeOR != NA) {

		if(mafCorrection == 0)
		    bias.push_back(L0[b][k].UdOR - L0[b][k].UeOR);
		else {
		    double maf = L0[b][k].snp->maf;
		    double factor = sqrt(2 * maf * (1-maf) );
		    bias.push_back((L0[b][k].UdOR - L0[b][k].UeOR) * factor);
		}

		Ud.push_back(L0[b][k].UdOR);
		Ue.push_back(L0[b][k].UeOR);
	    }


	if (bias.size() == 0)
	    biasAll.push_back(NA);
	else {
	    double factor = 1;

	    if(mafCorrection == 1) {
		double maf = Un[k].snp->maf;
		factor = sqrt(2 * maf * (1-maf));
	    }

	    biasAll.push_back(Un[k].odds - avg(bias) / factor );
	}

	if(level == 1) {
	    sprintf(s, "   SNP %3d:%s computed bias using %5d samples = Un[k]=%g Uboot=%g Bias=%g "
		"Bias.SD=%g Ud.SD=%g Ue.SD=%g\n", k, Un[k].snp->name.c_str(), (int) bias.size(),
		    Un[k].odds, biasAll[k], Un[k].odds - biasAll[k], sd(bias), sd(Ud), sd(Ue));

	    cout << s;
	}
    }

    return biasAll;
}


#include "plink-code/regression.h"


// Internal logistic regression code from plink
void fitLogistic(vector<Individual*> &persons, vector<Locus*> &locus, vector<LogisticOutput> &out)
{
    //vector<Locus*> locus = locus2;
    //sort(locus.begin(), locus.end() );
    out.clear();

    for(int i=0 ; i < locus.size() ; i ++)
    {
	LogisticRegression a;
	a.buildDesignMatrix(persons, locus[i]);
	a.fitLM();
	a.checkCollinearity();
	//a.displayResults(cout, locus[i]);
	out.push_back(a.getResults());
	//a.displayResults(outf, locus[i]);
    }


}

// Decided which logistic function to use and runs it
void logisticRegression(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs, vector<LogisticOutput> &results)
{
    if(useInternalRegression == 0) {
	plinkRunner.writeBinFiles(infile, Sample, SNPs, 1);
	plinkRunner.logistic(infile, results, SNPs);
    } else {
	fitLogistic(Sample, SNPs, results);
    }
}

/* mergeResults: Merge the logisticoutput results into the main results class */

void mergeResults(vector<Results> & out, vector<LogisticOutput> & UnLog)
{
    // plink output is sorted by CHR/BP so sort Un to be able to merge
    sort(out.begin(), out.end(), Results_sortByLocus);
    sort(UnLog.begin(), UnLog.end(), LogisticOutput_sortByLocus);

    //cout << "merge:\n";
    //for (uint i = 0; i < out.size(); i++) cout << out[i].snp->name << " "; cout << "\n";
    //for (uint i = 0; i < UnLog.size(); i++) cout << UnLog[i].snp->name << " "; cout << "\n";

    for (uint i = 0; i < UnLog.size(); i++) {
	out[i].odds = UnLog[i].odds;
	out[i].a1 = UnLog[i].a1;

	if (out[i].snp != UnLog[i].snp) {
	    cout << "Error: mergeResults: cannot merge results, snp's are not in expected order.\n";
	    cout << "Expected: " << out[i].snp->name << " Got: " << UnLog[i].snp->name << "\n";
	    gracefulExit(1);
	}

	if (0) cout << "SNP  " << out[i].snp->name << "\tOR=" << log(out[i].odds) << ","
		<< out[i].a1 << "\n";

	out[i].adjustNaive();
    }

    sort(out.begin(), out.end(), Results_sortByP);
}

void mergeResults(vector<Results> & out, vector<LogisticOutput> & Ud, vector<LogisticOutput> & Ue)
{
    // plink output is sorted by CHR/BP so sort Un to be able to merge
    sort(out.begin(), out.end(), Results_sortByLocus);
    sort(Ud.begin(), Ud.end(), LogisticOutput_sortByLocus);
    sort(Ue.begin(), Ue.end(), LogisticOutput_sortByLocus);

    if(0) {
    cout << "MERGE RESULTS: \n";
    for (uint i = 0; i < out.size(); i++) {
	char s[1000];
	sprintf(s, " %10s out %6.3f %6.3f %c Ud %6.3f %6.3f %c Ue %6.3f %6.3f %c\n",
	out[i].snp->name.c_str(),

	out[i].p, out[i].odds, 'U',
	Ud[i].p, Ud[i].odds, Ud[i].a1,
	Ue[i].p, Ue[i].odds, Ue[i].a1

	);
	cout << s;
    }
    }

    if (Ud.size() != Ue.size()) {
	cout << "Error: mergeResults: didn't get the same # of snps for D and E steps.\n";
	gracefulExit(1);
    }

    for (uint i = 0; i < out.size(); i++) {
	out[i].UdOR = Ud[i].odds;
	out[i].UdA1 = Ud[i].a1;

	out[i].UeOR = Ue[i].odds;
	out[i].UeA1 = Ue[i].a1;


	if (out[i].snp != Ud[i].snp || out[i].snp != Ue[i].snp) {
	    cout << "Error: mergeResults: cannot merge, "
		    "snp's are not in expected order.\n";

	    cout << "Expected: " << out[i].snp->name << " Got: " << Ud[i].snp->name << " "
		 << Ue[i].snp->name << "\n";

	    gracefulExit(1);
	}

	if (0) cout << "SNP  " << out[i].snp->name << "\tUd=" << log(out[i].UdOR) << ","
		<< out[i].UdA1 << "\tUe=" << log(out[i].UeOR) << "," << out[i].UeA1 << "\n";

	out[i].adjust();

	out[i].odds = out[i].UdOR;
	out[i].a1 = out[i].UdA1;
    }

    sort(out.begin(), out.end(), Results_sortByP);
}

void testQTAssociation(vector<Individual*> &Sample, vector<Locus*> &SNPs,
	vector<LogisticOutput> &results, const string &prefix)
{
    if(usePlinkCode == 0 || haveCovariates == 1) {
	stable_sort(SNPs.begin(), SNPs.end(), Locus_sortByCHR);
	plinkRunner.writeBinFiles(prefix, Sample, SNPs, 1);
	plinkRunner.linear(prefix, results, SNPs);
    }
    else {
	plinkInternalCode.testQAssoc(0, Sample, SNPs, results);
    }
}



/*
 *  Use basic QT association to get SNP p-values
 */

int isSignificant_QTL(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
	int newfids, vector<ModelOutput> &out)
{
    // Assume testType == QTLASSOC

    vector<LogisticOutput> results;

    testQTAssociation(Sample, SNPs, results, infile);

    // Convert results to ModelOutput since we want to return a vector of ModelOutput
    out.clear();
    ModelOutput temp;

    for (int i = 0; i < results.size(); i++) {
	temp.snp = results[i].snp;
	temp.p = results[i].p;
	out.push_back(temp);
    }

    int i;
    int numSig = 0;

    for (i = 0; i < out.size(); i++) {
	ModelOutput *cur = &out[i];
	cur->significant = 0;
	if (cur->p != NA && cur->p <= alpha) {
	    cur->significant = 1;
	    numSig++;
	}
    }

    if (debug == 1) cout << "isSignificant_QTL: infile=" << infile << " out.size()= " << out.size()
	    << " numSig = " << numSig << "\n";

    return numSig;
}



/*
 *  Use basic allelic test to get SNP p-values
 */

int isSignificant_Allelic(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
	int newfids, vector<ModelOutput> &out)
{
    vector<LogisticOutput> results;

    plinkInternalCode.testAssoc(Sample, SNPs, results);

    // Convert results to ModelOutput since we want to return a vector of ModelOutput
    out.clear();
    ModelOutput temp;

    for (int i = 0; i < results.size(); i++) {
	temp.snp = results[i].snp;
	temp.p = results[i].p;
	out.push_back(temp);
	//cout << "xxx " << temp.p << "\n";
    }

    int i;
    int numSig = 0;

    for (i = 0; i < out.size(); i++) {
	ModelOutput *cur = &out[i];
	cur->significant = 0;
	if (cur->p != NA && cur->p <= alpha) {
	    cur->significant = 1;
	    numSig++;
	}
    }

    if (debug == 1) cout << "isSignificant_Allelic: infile=" << infile << " out.size()= " << out.size()
	    << " numSig = " << numSig << "\n";

    return numSig;
}


/*
 *  Use R plugin to test for association
 */

int isSignificant_R(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
			  int newfids, vector<ModelOutput> &out)
{
    vector<LogisticOutput> results;
    
    stable_sort(SNPs.begin(), SNPs.end(), Locus_sortByCHR);
    plinkRunner.writeBinFiles(infile, Sample, SNPs, 1);
    plinkRunner.runRplugin(infile, results, SNPs, RScriptFile);
    
    // Convert results to ModelOutput since we want to return a vector of ModelOutput
    out.clear();
    ModelOutput temp;
    
    int numSig;
    
    for (int i = 0; i < results.size(); i++) {
	temp.snp = results[i].snp;
	temp.p = results[i].p;
	out.push_back(temp);
    }

    return numSig;
}


// Set to 1 if we use screening by trend test to speed-up estimation of p-values using logistic test
int screenWithTrendTest = 1;

// The p-value for the trend test that is regarded as significant.
double screenPValue = NA;

/*
 *  Use logistic test to get SNP p-values
 */

int isSignificant_Logistic(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
	int newfids, vector<ModelOutput> &out)
{
    vector<LogisticOutput> results;
    vector<Locus*> screenedSNPs;

    // Estimate screening p-value for trend test
    if(screenWithTrendTest == 1 && screenPValue == NA) {

	cout << "Calculating screening trend test p-value: ";
	vector<ModelOutput> out;
	plinkInternalCode.fullModelAssoc(0, Sample, SNPs, out);

	logisticRegression(infile, Sample, SNPs, results);

	char s[1000];

	#pragma omp critical
	{
	// ofstream file("screen.txt", ios::app);

	double P = 0;

	for (int i = 0; i < out.size(); i++) {
	    ModelOutput &cur = out[i];
	    LogisticOutput &log = results[i];

	    if(log.p > alpha ) continue;

	    sprintf(s,"%10s %10s %4g %4g\n", cur.snp->name.c_str(), log.snp->name.c_str(),
		    cur.pTrend, log.p);

	    if(log.p < alpha) if(P < cur.pTrend) P = cur.pTrend;

	    // file << s;
	}

	cout << "P=" << P << " alpha=" << alpha << " ";
	screenPValue = P * 10;

	cout << "screen P = " << screenPValue << "\n";

	}
    }

    // Assume testType == LOGISTIC
    if(screenPValue != NA) {
	vector<ModelOutput> out;
	plinkInternalCode.fullModelAssoc(0, Sample, SNPs, out);

	for (int i = 0; i < out.size(); i++) {
	    ModelOutput &cur = out[i];
	    if(cur.pTrend < screenPValue) screenedSNPs.push_back(cur.snp);
	}

	cout << " Screened Snps: " << screenedSNPs.size() << " out of " << SNPs.size() <<
	"\tP = " << screenPValue << "\n";

	logisticRegression(infile, Sample, screenedSNPs, results);
    } else {
	logisticRegression(infile, Sample, SNPs, results);
    }

    // Convert results to ModelOutput since we want to return a vector of ModelOutput
    out.clear();
    ModelOutput temp;

    for (int i = 0; i < results.size(); i++) {
	temp.snp = results[i].snp;
	temp.p = results[i].p;
	out.push_back(temp);
    }

    int numSig = 0;

    for (int i = 0; i < out.size(); i++) {
	ModelOutput &cur = out[i];
	cur.significant = 0;
	if (cur.p != NA && cur.p <= alpha) {
	    cur.significant = 1;
	    numSig++;
	}
    }

    if (debug == 1) cout << "isSignificant_logistic: infile=" << infile << " out.size()= "
	    << out.size() << " numSig = " << numSig << "\n";

    return numSig;
}

int isSignificant_Model(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
		  int newfids, vector<ModelOutput> &out)
{
    // Use association code from plink
    if (usePlinkCode == 1) {
	// No need for critical section now, but make sure copied plink code is thread safe
	plinkInternalCode.fullModelAssoc(0, Sample, SNPs, out);
    } else {
	plinkRunner.writeBinFiles(infile, Sample, SNPs, newfids);
	plinkRunner.model(infile, infile, out);
    }
    
    unsigned int i;
    int numSig = 0;
    
    // Tag significant SNPs
    
    for (i = 0; i < out.size(); i++) {
	ModelOutput *cur = &out[i];
	
	cur->significant = 0;
	
	// if test is minTrendGeno then use min p value of trend and geno;
	if (testType == minTRENDGENO) {
	    cur->p = min(cur->pGeno, cur->pTrend);
	}
	
	if (testType == GENO) cur->p = cur->pGeno;
	
	if (testType == TREND) cur->p = cur->pTrend;
	
	if (cur->p != NA && cur->p <= alpha) {
	    cur->significant = 1;
	    numSig++;
	}
	
	// cout << cur->snp->name << " " << cur->pGeno << " " << cur->pTrend << "\n";
	// printf("SIG: %s %f %f \n", cur->snp->name.c_str(), cur->pGeno, cur->pTrend);
    }
    
    if (debug == 1) cout << "isSignificant: infile=" << infile << " out.size()= " << out.size()
	<< " numSig = " << numSig << "\n";
    
    return numSig;
}

/*
 *  Run the appropriate test and tag significant SNPS
 */

int isSignificant(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
	int newfids, vector<ModelOutput> &out)
{
    // Run the appropriate significance test

    if (QuantTrait == 1) return isSignificant_QTL(infile, Sample, SNPs, newfids, out);

    if (testType == ALLELIC) return isSignificant_Allelic(infile, Sample, SNPs, newfids, out);

    if (testType == LOGISTIC) return isSignificant_Logistic(infile, Sample, SNPs, newfids, out);
    
    if (testType == RPLUGIN) return isSignificant_R(infile, Sample, SNPs, newfids, out);

    if (testType == DUMMY) return DummyDataTests::detection(infile, Sample, SNPs, newfids, out);

    return isSignificant_Model(infile, Sample, SNPs, newfids, out);
}



// Get OR for selection sample and SNPs

void getOR(vector<Individual*> &sample, vector<Locus*> &SNPs, const string &prefix,
		vector<LogisticOutput> &out)
{
    if (estimationType == DUMMY_MODEL) {
	DummyDataTests::estimation(prefix, prefix, out, SNPs);
    }
    else if (estimationType == ALLELIC_MODEL) {
	plinkInternalCode.testAssoc(sample, SNPs, out);
    }
    else if (estimationType == RPLUGIN_MODEL) {
	stable_sort(SNPs.begin(), SNPs.end(), Locus_sortByCHR);
	plinkRunner.writeBinFiles(prefix, sample, SNPs, 1);
	plinkRunner.runRplugin(prefix, out, SNPs, RScriptFile);
    }
    else if (QuantTrait == 1) {
	testQTAssociation(sample, SNPs, out, prefix);
    } else {
	// plink will output these sorted, so we'd better sort them now otherwise reading will be harder
	stable_sort(SNPs.begin(), SNPs.end(), Locus_sortByCHR);
	logisticRegression(prefix, sample, SNPs, out);
    }
}


/*
 * getBias: computes Ud_Ue component of bias for a bootstrap sample
 *
 */
int getBias(vector<Results> & out, BootStrap &BS, vector<Locus*> & SNPs, int maxSNPsToProcess,
	char *id, char *thread_id, int level)
{
    char s[1000];
    string prefix = tmpdir + "temp_" + thread_id;

    // Check if bootstrap sample is significant

    vector<ModelOutput> model;
    int numSig = isSignificant(prefix + "all", BS.BootstrapSample, SNPs, 1, model);

    if(level != 2) {
        sprintf(s, "Level %s:  significant snps=%d \n", id, numSig);
        cout << s;
    }

    //outputPvalues((string)"p" + (string)id, model);

    if (numSig == 0) return 0;

    // Sample is significant, compute Ud/Ue quantities

    // Sort model results by p value
    sort(model.begin(), model.end());

    if (debug) {
    #pragma omp critical
	{
	    for (int i = 0; i < model.size(); i++) {
		model[i].snp -> ranks.push_back(i + 1);
		if (model[i].significant) model[i].snp->numSig ++;
	    }

	    totalSamples++;
	}
    }

    // Select only significant SNPs for analysis

    vector<Locus*> selected;
    out.clear();
    out.reserve(numSig);

    int sig = 0;

    for (unsigned int i = 0; i < model.size(); i++)
	if (model[i].significant) {
	    selected.push_back(model[i].snp);
	    Results r(model[i].snp, model[i].p);
	    out.push_back(r);
	    sig++;

	    // Break if we exceed maxSNPsToProcess
	    if (maxSNPsToProcess != -1 && sig >= maxSNPsToProcess) break;
	}

    stable_sort(selected.begin(), selected.end(), Locus_sortByCHR); // Sort by CHR/BP

    vector<LogisticOutput> Ud, Ue;

    string udfile = prefix + "ud";
    string uefile = prefix + "ue";

    /* Compute Ud/Ue */

    getOR(BS.BootstrapSample, selected, udfile, Ud);
    getOR(BS.NotSelected, selected, uefile, Ue);
    mergeResults(out, Ud, Ue);
    return out.size();
}

/*
 * getNaiveEstimate():
 *   Compute Naive estimate Un from original dataset.
 *
 *   Returns the SNPs selected for further analysis
 *
 */

vector<Locus*> getNaiveEstimate()
{
    vector<ModelOutput> UnSig;
    char s[1024]; // temporary storage for printing

    cout << "Computing naive estimate...\n\n";

    // Check significant SNPs in original dataset
    Sn = isSignificant(tmpdir + "temp_all", persons, locus, 0, UnSig);

    if (Sn == 0) {
	cout << "\nNo Significant SNPs in input file. Nothing to do.\n";
	gracefulExit(1);
    }

    sort(UnSig.begin(), UnSig.end()); // sort by p-value

    for (int i = 0; i < UnSig.size(); i++)
	UnSig[i].snp->rank = i + 1;

    for (int i = 0; i < UnSig.size(); i++)
	UnSig[i].snp->p = UnSig[i].p;

    sprintf(s, "Original dataset: # of significant snps = %d \n", Sn);
    cout << s;

    // Select only top <k> SNPs for analysis
    // TODO: this should be re-implemented as changing Sn has some side effects

    if (maxSNPs != -1) { // option --top specified
	cout << "Selecting only the top " << maxSNPs << " significant SNPs\n";

	int sig = 0;
	for (uint i = 0; i < UnSig.size(); i++)
	    if (UnSig[i].significant) {
		if (sig + 1 > maxSNPs) {
		    UnSig[i].significant = 0;
		    continue;
		}

		sig++;
	    }

	Sn = sig;
    }

    // Keep only top <k> significant SNPs for bootstrapping and delete all others
    if (keeptop != -1) { // option --keep specified
	cout << "Option --keep specified: Keeping only " << keeptop << " SNPs to speed up computation.\n";

	if (keeptop < Sn) {
	    cout << "Error: Cannot keep less than the number of significant SNPs\n";
	    gracefulExit(1);
	}

	for (uint i = keeptop; i < UnSig.size(); i++)
	    UnSig[i].snp->id = -1;

	deleteSNPS();

	cout << "Number of snps now is " << locus.size() << "\n";

	// Now run function again
	keeptop = -1;
	return getNaiveEstimate();
    }

    // Keep only the significant and a random number of snps (for debugging purposes).

    if (keeprandom != -1) { // option --keep specified
	cout << "Warning: Option --keeprandom specified: Keeping only " << keeprandom << " SNPs.\n";

	vector<Locus*> sigs, notsigs;

	// Split SNPs
	for(int i = 0; i < UnSig.size(); i++)
	    if(UnSig[i].significant)
		sigs.push_back( UnSig[i].snp );
	    else
		notsigs.push_back( UnSig[i].snp );

	cout << sigs.size() << " significant and " << notsigs.size() << " non-significant SNPS\n";

	locus.resize(0);
	for(int i = 0; i < sigs.size(); i++)
	    locus.push_back(sigs[i]);

	if(keeprandom > notsigs.size()) keeprandom = notsigs.size();

	// Add the random snps
	for(int i = 0; i < keeprandom; i++)
	{
	    int p = 0;

	    p = RNG.GetRand(0, notsigs.size()-1);
	    while(notsigs[p]-> chr == 99)
		p = RNG.GetRand(0, notsigs.size()-1);

	    locus.push_back(notsigs[p]);
	    notsigs[p]->chr = 99;
	}

	//deleteSNPS();

	cout << "SNPs remaining now are " << locus.size() << "\n";
	nlocus = locus.size();

	// Now run function again
	keeprandom = -1;
	return getNaiveEstimate();
    }



    if (debug >= 1) {
	sprintf(s, "Significant SNPs:\n");
	cout << s;

	for (uint i = 0; i < UnSig.size(); i++)
	    if (UnSig[i].significant) {
		sprintf(s, "%s(%g) ", UnSig[i].snp->name.c_str(), UnSig[i].p);
		cout << s;
	    }

	sprintf(s, "\n\n");
	cout << s;
    }

    // Select only signigicant SNPs for analysis
    vector<Locus*> selected;
    for (unsigned int i = 0; i < UnSig.size(); i++)
	if (UnSig[i].significant) {
	    selected.push_back(UnSig[i].snp);
	    Results r(UnSig[i].snp, UnSig[i].p);
	    Un.push_back(r);
	}

    return selected;
}

/*
 * Function getNaiveOR:
 *  run plink and get original Un estimate;
 *
 *  persons: Which persons to use
 *  selectedSNPs: Which SNPs to use
 *
 */

void getNaiveOR(vector<Individual*> & persons, vector<Locus*> &selectedSNPS)
{
    vector<LogisticOutput> UnLog;

    // Get odds ratio for these SNPs
    getOR(persons, selectedSNPS, tmpdir + "temp_" + "Un", UnLog);

    // Merge plink results into main matrix Un
    mergeResults(Un, UnLog);

    if (silent == 0) {
	char s[100];
	cout << "Naive p-values and OR:\n";
	for (uint i = 0; i < Un.size(); i++) {
	    sprintf(s, "%s %g %g (%c)\n", Un[i].snp->name.c_str(), Un[i].p, pow(M_E, Un[i].odds),
		    Un[i].a1);
	    cout << s;
	}
	cout << "\n\n";
    }
}

/* 
 * function printBias:
 *   Prints a matrix of bias results
 *
 */

void printBias(const string &label, const string &prefix, vector<Results> &v)
{
    unsigned int n = v.size();
    char s[1000];
    if (silent == 1) return;
    cout << "Matrix " << label << "\n";

    for (unsigned int i = 0; i < n; i++) {
	sprintf(s, "%s %4d: %15s p=%12g   UdOR,A1=(%4.3f,%c)   UeOR,A1=(%4.3f,%c)\n",
		prefix.c_str(), i, v[i].snp->name.c_str(), v[i].p,
		pow(M_E, v[i].UdOR), v[i].UdA1,
		//v[i].UdOR, v[i].UdA1,
		pow(M_E, v[i].UeOR), v[i].UeA1);
		//v[i].UeOR, v[i].UeA1);
	cout << s;
    }
}

vector<double> Uboot0; // Bias reduced OR estimate, computed from level 0 quantities


/* Level 1/2 Quantities */

vector<int> S1; // number of significant SNPs in each iteration of level 1
vector<vector<Results> > L1;

vector<vector<int> > S2; // number of significant SNPs in each iteration of level 2
vector<vector<vector<Results> > > L2;

vector<vector<double> > Uboot; // Bias estimates from level 1 and 2 (used to compute SD).

/** Std Deviation as computed by L2 bootstrap */
vector<double> StdDev;

/** Reliability index is percent of entries used to compute CI of a SNP */
vector<double> RI;

/* Compute Level 0 Quantities */
void getLevel0Quantities()
{
    vector<int> S0;
    vector<vector<Results> > L0;

    int N = 0;
    int m = 0;
    
    cout << "Computing Level 0 quantities.\n\n";

    // used to compute estimated time left
    time_t start = time(NULL);

    L0.reserve(Level0Bootstraps + 10);
    S0.reserve(Level0Bootstraps + 10);

    int iteration = 0;

    // The max number of iterations is 10*specified number, to avoid running indefinetely
    // in case we can't get significant bootstraps

    // loop has been parallelized using OpenMP

    // It is not possible to break from a loop when using openmp
    // So we run 10*bootstraps and when we get enough significant ones
    // we just do continue until loop is finished

    #pragma omp parallel for shared(m)
    for (N = 0; N < 10 * Level0Bootstraps; N++) {
	if (m >= Level0Bootstraps) continue; // skip if we got enough bootstraps

	char id[100] = "", thread_id[100] = "";
	int myIteration;

	BootStrap L0BS(&persons);

	// This has to be critical since the random number generator is not re-entrant

	#pragma omp critical
	{
	    L0BS.generateBootstrap();
	    myIteration = iteration++; // get an iteration id
	}

	#ifdef _OPENMP
	//if(silent == 0) printf("Thread: %d N=%d m=%d L0=%d\n",omp_get_thread_num(), N, m, L0.size());
	sprintf(thread_id,"%d", omp_get_thread_num());
	#endif

	sprintf(id, "L0-%d", myIteration);

	vector<Results> out;
	int numSig = getBias(out, L0BS, locus, Sn, id, thread_id, 0);
	if (numSig == 0) continue;

	// Add entries to S0 and L0
	#pragma omp critical
	{
	    L0.push_back(out);
	    S0.push_back(numSig);

	    if (numSig >= Sn) m++;

	    printBias("L0[N]", "  ", out);
	    cout.flush();
	}

	/* Computing of time estimate. This won't work in multi-threaded code so is left out for now */
	if (0 && m == 2) {
	    char s[500];
	    time_t now = time(NULL);
	    double time_est = Level0Bootstraps * ((double) now - start + 1) / m;

	    if (time_est > 100) sprintf(s, "%.1f min", time_est / 60.0);
	    else sprintf(s, "%f sec", time_est);

	    cerr << "Estimated time to finish L0: " << s << " (" << ((int) (59 + time_est - (now
		    - start)) / 60) << " min left)\n";

	}

    }

    // cout << m << " " << iteration << "\n";
    cout << "Level 0 completed. Significant Bootstraps were " << ((double) m / iteration) * 100
	    << "% \n";

    // If we are running on a cluster stop here and output partial results
    if( clusterN > 0) {
        outputL0(clusterFile, L0);
        return;
    }

    N = L0.size();

    // Compute Uboot0[k] k=1..Sn

    Uboot0 = computeBias(Un, L0, 1);

    // Truncation of Uboot at 0
    if (!notruncate) for (int i = 0; i < Sn; i++)
	if (Uboot0[i] != NA && Uboot0[i] < 0) Uboot0[i] = 0;

    // Print all ud and ue samples if user wants
    if (logAllBias == 1) {
	ofstream out("all-bias.txt");
	ofstream out_ud("ud-bias.txt");
	ofstream out_ue("ue-bias.txt");

	for (int k = 0; k < Sn; k++) out << Un[k].snp->name << " ";
	out << "\n";
	for (int k = 0; k < Sn; k++) out_ud << Un[k].snp->name << " ";
	out_ud << "\n";
	for (int k = 0; k < Sn; k++) out_ue << Un[k].snp->name << " ";
	out_ue << "\n";

	for (int b = 0; b < N; b++) {
	    for (int k = 0; k < Sn; k++) { // For each significant snp
		vector<double> bias;

		if (L0[b].size() > k && L0[b][k].UdOR != NA && L0[b][k].UeOR != NA) {
		    char s[200];
		    sprintf(s, "%.6f ", (L0[b][k].UdOR - L0[b][k].UeOR));
		    out << s;
		} else out << "      NA ";

		if (L0[b].size() > k && L0[b][k].UdOR != NA) {
		    char s[200];
		    sprintf(s, "%.6f ", (L0[b][k].UdOR));
		    out_ud << s;
		} else out_ud << "      NA ";

		if (L0[b].size() > k && L0[b][k].UeOR != NA) {
		    char s[200];
		    sprintf(s, "%.6f ", (L0[b][k].UeOR));
		    out_ue << s;
		} else out_ue << "      NA ";

	    }
	    out << "\n";
	    out_ud << "\n";
	    out_ue << "\n";
	}
    }

    cout.flush();
}

/*
 * Compute level 1 and level 2 quantities
 */

void getLevel12Quantities()
{
    int N = 0; // number of significant level 1 iterations completed so far
    int Nrep; // number of iterations completed so far

    // used to compute time estimate to finish
    time_t start = time(NULL);

    cout << "\nComputing level 1 & 2 quantities...\n";

    int iteration = 0;

    #pragma omp parallel for
    for (Nrep = 0; Nrep < 5 * Level1Bootstraps; Nrep++) {
	char s[500], id[100] = "", thread_id[100] = "";
	int myIteration;

	if (N >= Level1Bootstraps) continue;

	BootStrap L1BS(&persons);

	// This has to be critical since the random number generator is not reentrant.

	#pragma omp critical
	{
	    L1BS.generateBootstrap();
	    myIteration = iteration++;
	}

	sprintf(id, "L1-%d", myIteration);

	int myS1;
	vector<Results> myL1;
	vector<int> myS2;
	vector<vector<Results> > myL2;

	#ifdef _OPENMP
	sprintf(thread_id,"%d", omp_get_thread_num());
	#endif

	int numSig = getBias(myL1, L1BS, locus, Sn, id, thread_id, 1);
	//int numSig = getOR(myL1, L1BS.BootstrapSample, locus, Sn, id, thread_id);

	// Only store significant bootstraps
	if (numSig == 0) continue;

	// Add entries to S1 and L1
	myS1 = numSig;

	printBias("L1[N]", "    ", myL1);

	/*
	 *
	 *  LEVEL 2 Bootstrap
	 *
	 *
	 */

	BootStrap L2BS(&L1BS.BootstrapSample);

	myL2.resize(Level2Bootstraps);
	myS2.resize(Level2Bootstraps);

	vector<Results> out;

	int M = 0;
	while (M < Level2Bootstraps) {

	    // Break fromp openmp loop if we already got enough bootstraps
	    if (N >= Level1Bootstraps) {
		M++;
		continue;
	    }

	    char id[100] = "";
	    sprintf(id, "L2_%d-%d", myIteration, M);

	    #pragma omp critical
	    {
		L2BS.generateBootstrap();
	    }

	    // Check if bootstrap sample is significant
	    int numSig = getBias(out, L2BS, locus, myS1, id, thread_id, 2);

	    // sprintf(s,"L2: S2[%d][%d] = %d\n",N,M, numSig); cout << s;

	    if (numSig == 0) continue;

	    myL2[M] = out;
	    myS2[M] = myL2[M].size();

	    printBias("L2[N][M]", "       ", myL2[M]);

	    M++;
	}

	// Compute time estimate
	if (N > 1) {
	    char s[500];
	    time_t now = time(NULL);
	    double time_est = Level1Bootstraps * ((double) now - start + 0.5) / N;

	    if (time_est > 100)
		sprintf(s, "%.1f min (%.0f%% done)",
		    (time_est - (now- start)) / 60,
		    100 * (double)N / Level1Bootstraps
		);
	    else sprintf(s, "%.1f sec (%.0f%% done)",
		    (time_est - (now- start)),
		    100 * (double)N / Level1Bootstraps
		 );

	    cerr << "Estimated time to complete analysis: " << s << "\n";
	    cout.flush();
	}

	if (N >= Level1Bootstraps) continue;

	vector<double> bias = computeBias(myL1, myL2, 2);

	// Completed a level 1 iteration, now add data back to global matrices L1,L2
	#pragma omp critical
	{
	    Uboot.push_back(bias);
	    S1.push_back(myS1);

	    N++;
	}

    }

    // If we are running on a cluster stop here and output partial results
    if( clusterN > 0) {
        outputL1(clusterFile);
        return;
    }

}

/**
 *
 * computeFinalResults(): Compute Uboot[k][i] , confidence intervals and reliability rating
 *
 */

void computeFinalResults()
{
    if (debug >= 2) {
	cout << "Matrix Sizes:\n ";

	cout << "\n\nL1:";
	for (int i = 0; i < L1.size(); i++) cout << L1[i].size() << " ";
	cout << "\n";

	cout << "\n\nS1:";
	for (int i = 0; i < S1.size(); i++) cout << S1[i] << " ";
	cout << "\n";

	cout << "\nL2: \n";
	for (int i = 0; i < L2.size(); i++) {
	    for (int j = 0; j < L2[i].size(); j++)
		cout << L2[i][j].size() << " ";

	    cout << "\n";
	}
	cout << "\n\n";
    }

    int n1 = Uboot.size(); // number of computed samples at level 1

    // Uboot[bootstrapnum][snp]
    // Uboot[i][k] = UdL1[i][k] - avg over j( UdL2[i][j][k] - UeL2[i][j][k] )

    // This now happens inside the main loop to reduce memory requirements:
    //for (int i = 0; i < n1; i++) {
    //	Uboot.push_back(computeBias(L1[i], L2[i], 2));
    //}


    // StdDev[i] = mean over row k of (Uboot_k^i)


    if (debug) {
	cout << "UBOOT SIZE: " << Uboot.size() << " ";
	if(Uboot.size() > 0 ) cout << Uboot[0].size() << "\n"; else cout << " NA \n";

	for(uint k = 0; k < Uboot.size(); k++) {
	    cout << "Uboot[ " << k << "] : ";
	    for (int i = 0; i < Uboot[k].size(); i++)
		cout << toString(Uboot[k][i], "%8.2g") << " ";
	    cout << "\n";
	}
    }

    // Compute SD for each SNP
    StdDev.resize((int) Sn);

    for (uint k = 0; k < Sn; k++) {
	// compute stddev
	vector<double> all;

	for (uint i = 0; i < Uboot.size(); i++)
	    if (Uboot[i].size() > k && Uboot[i][k] != NA) all.push_back(Uboot[i][k]);

	StdDev[k] = sd(all);

	if (debug) {
	    cout << "Snp " << k << " computed sd using " << all.size()
	         << " samples = " << sd(all) << "\n";

	    cout << "SD: ";
	    for (uint i = 0; i < all.size(); i++)
		cout << all[i] << " ";
	    cout << "\n";
	}
    }



    // Compute Reliability (i.e. percentage of samples where used in the stddev calculation

    RI.resize((int) Sn);
    for (int k = 0; k < Sn; k++)
	RI[k] = 0;

    for (uint i = 0; i < L2.size(); i++) {
	for (uint j = 0; j < L2[i].size(); j++) {
	    int n = L2[i][j].size();
	    if (n > Sn) n = Sn;

	    for (int k = 0; k < n; k++)
		RI[k]++;
	}
    }

    if (n1 * Level2Bootstraps != 0) for (int k = 0; k < Sn; k++)
	RI[k] /= (n1 * Level2Bootstraps);
    else for (int k = 0; k < Sn; k++)
	RI[k] = NA;
}


#if 0
void computeFinalResults_old()
{
    char s[1000];
    vector<vector<double> > Uboot;

    if (debug) {
	cout << "Matrix Sizes:\n ";

	cout << "\n\nL1:";
	for (int i = 0; i < L1.size(); i++)
	    cout << L1[i].size() << "," << S1[i] << " ";
	cout << "\n";

	cout << "\nL2: \n";
	for (int i = 0; i < L2.size(); i++) {
	    for (int j = 0; j < L2[i].size(); j++)
		cout << L2[i][j].size() << " ";

	    cout << "\n";
	}
	cout << "\n\n";
    }

    int n1 = S1.size(); // number of computed samples at level 1

    Uboot.resize(n1);

    // Uboot[bootstrapnum][snp]
    // Uboot[i][k] = UdL1[i][k] - avg over j( UdL2[i][j][k] - UeL2[i][j][k] )

    for (int i = 0; i < n1; i++) {

	Uboot[i].resize(S1[i]);

	for (int k = 0; k < S1[i]; k++) {

	    Uboot[i][k] = 0;

	    double avg = 0;
	    int n = 0;

	    // Compute mean over row k of (UD i_j_k - Ue i_j_K )
		cout << "SNP" << k << " : ";

	    for (int j = 0; j < Level2Bootstraps; j++) {

		if (L2[i][j].size() > k) if (L2[i][j][k].UdOR != NA && L2[i][j][k].UeOR != NA) {
		    avg += L2[i][j][k].UdOR - L2[i][j][k].UeOR;
		    n++;

		    cout << " Ud=" + toString(L2[i][j][k].UdOR, "%5.2g") + " Ue=" +toString(L2[i][j][k].UeOR, "%5.2g");

		}
	    }
		cout << "\n";

	    avg = avg / n;

	    /*(if(L1[i][k].UdOR < 1) {
	     L1[i][k].UdOR = 1 / L1[i][k].UdOR;
	     }*/

	    Uboot[i][k] = L1[i][k].UdOR - avg;

	    sprintf(s, "   Average for SNP %3d:%s using %d samples = Un[k]=%g,%g Uboot=%g "
	    	    " \n", k, L1[i][k].snp->name.c_str(), n,
	    		L1[i][k].odds, L1[i][k].UdOR, Uboot[i][k]);

	    	cout << s;

	}
    }

    // Compute SD for each SNP
    StdDev.resize((int) Sn);



    if (debug) {
	cout << "UBOOT SIZE: " << Uboot.size() << " " << Uboot[0].size() << "\n";

	for(uint k = 0; k < Uboot.size(); k++) {
	    cout << "Uboot[ " << k << "] : ";
	    for (int i = 0; i < Uboot[k].size(); i++)
		cout << toString(Uboot[k][i], "%8.2g") << " ";
	    cout << "\n";
	}
    }

    // StdDev[i] = mean over row k of (Uboot_k^i)

    for (uint k = 0; k < Sn; k++) {
	// compute stddev
	vector<double> all;

	for (uint i = 0; i < n1; i++)
	    if (Uboot[i].size() > k && Uboot[i][k] != NA) all.push_back(Uboot[i][k]);

	if (debug) {
	    cout << "Uboot[ " << k << "] : ";
	    for (int i = 0; i < all.size(); i++) {
		cout << toString(all[i], "%.3g") << " ";
	    }
	    cout << "\n";
	}

	StdDev[k] = sd(all);

	if (debug) {
	    cout << "Snp " << k << " computed sd using " << all.size() << " samples = " << sd(all)
		    << "\n";
	}

	if (debug) {
	    cout << "SD: ";
	    for (uint i = 0; i < all.size(); i++)
		cout << all[i] << " ";
	    cout << "\n";
	}
    }

    // Compute Reliability (i.e. percentage of samples where used in the stddev calculation

    RI.resize((int) Sn);
    for (int k = 0; k < Sn; k++)
	RI[k] = 0;

    for (uint i = 0; i < L2.size(); i++) {
	for (uint j = 0; j < L2[i].size(); j++) {
	    int n = L2[i][j].size();
	    if (n > Sn) n = Sn;

	    for (int k = 0; k < n; k++)
		RI[k]++;
	}
    }

    if (n1 * Level2Bootstraps != 0) for (int k = 0; k < Sn; k++)
	RI[k] /= (n1 * Level2Bootstraps);
    else for (int k = 0; k < Sn; k++)
	RI[k] = NA;

}
#endif

/*
 * Output the resulting table
 */
void outputFinalResults()
{
    char s[1000], CI_str[1000];

    if(debug) {
	ofstream psig("psig.txt");

	vector<Locus *> a = locus;
	sort(a.begin(), a.end(), Locus_sortbyP );

	for (int i = 0; i < a.size(); i++) {
	    char s[200];
	    sprintf(s, "%10s %10.3g %.4f\n", a[i]->name.c_str(), a[i]->p, ( a[i]->numSig / (double)totalSamples ));
	    psig << s;


	}
    }


    vector<string> d;
    vector<vector<string> > data;

    cout << "\n\nAnalysis finished.\n";
    cout << "There were " << Sn << " significant SNPs in original dataset.\n";

    ofstream out;
    out.open(resultsFile.c_str(), ios::out);

    d.resize(100);

    int p = 0;

    d[p++] = "CHR";
    d[p++] = "   SNP";
    d[p++] = " BP";
    d[p++] = " MAF";

    //d[p++] = " MAFCONT";
    //d[p++] = " MAFCASE";
    d[p++] = "A1";

    d[p++] = "        P";

    if (QuantTrait == 1)
	d[p++] = "  BETA";
    else {
	if(reportLogOR == 1) d[p++] = "  LOGOR"; else d[p++] = "     OR";
    }


    d[p++] = "  UBOOT";
    d[p++] = "SD";
    d[p++] = "CI";
    //d[p++] = "  RI";

    if(debug) d[p++] = "PSIG";

    d.resize(p);
    data.push_back(d);

    for (int i = 0; i < Sn; i++) {

	p = 0;

	d[p++] = toString(Un[i].snp->chr);
	d[p++] = Un[i].snp->name;
	d[p++] = toString(Un[i].snp->bp);
	d[p++] = toString(Un[i].snp->maf, "%.3f");

	//d[p++] = toString(Un[i].snp->mafControls,"%.3f");
	//d[p++] = toString(Un[i].snp->mafCases,"%.3f");
	d[p++] = Un[i].a1;

	d[p++] = toString(Un[i].p, "%.3g");

	if(reportLogOR == 1 || QuantTrait == 1)
	{
	    d[p++] = toString(Un[i].odds, "%.4f");
	    d[p++] = toString(Uboot0[i], "%.4f");
	    d[p++] = toString(StdDev[i], "%.4f");

	    if (StdDev[i] == NA) sprintf(CI_str, "(%s,%s)", "NA", "NA");
	    else sprintf(CI_str, "(%.3f,%.3f)", Uboot0[i] - 1.96 * StdDev[i], Uboot0[i] + 1.96 * StdDev[i]);

	    d[p++] = CI_str;
	}
	else
	{
	    d[p++] = toString(exp(Un[i].odds), "%.4f");
	    d[p++] = Uboot0[i] != NA ? toString(exp(Uboot0[i]), "%.4f") : "NA";
	    d[p++] = StdDev[i] != NA ? toString(exp(StdDev[i]), "%.4f") : "NA";

	    if (StdDev[i] == NA) sprintf(CI_str, "(%s,%s)", "NA", "NA");
	    else sprintf(CI_str, "(%.3f,%.3f)", exp(Uboot0[i] - 1.96 * StdDev[i]), exp(Uboot0[i] + 1.96 * StdDev[i]));

	    d[p++] = CI_str;
	}


	// d[p++] = toString(RI[i], "%g");
	if(debug) d[p++] = toString( Un[i].snp->numSig / (double)totalSamples, "%.3f");

	data.push_back(d);
    }

    printTable(cout, data);
    printTable(out, data);
}








void outputPvalues(string file, vector<ModelOutput> out)
{
    uint i;
    ofstream f;
    vector<double> v;
    f.open(file.c_str(), ios::out);

    for (i = 0; i < out.size(); i++) {

	f << i << " " << out[i].p << "\n";
	v.push_back(out[i].p);
    }

    histogram(v, 200, file + ".hist");
}

void printStats()
{
    vector<string> header(10);
    vector<vector<string> > data;

    cout << "Writing statistics to file 'br2.stats' .\n";

    ofstream out;
    out.open("br2.stats", ios::out);

    header.resize(8);
    header[0] = "CHR";
    header[1] = "      SNP";
    header[2] = " BP";
    header[3] = "A1";
    header[4] = "A2";
    header[5] = "MAF";
    header[6] = "MAFCases";
    header[7] = "MAFControls";

    data.push_back(header);

    for (int i = 0; i < locus.size(); i++) {
	vector<string> d(8);

	d[0] = toString(locus[i]->chr);
	d[1] = locus[i]->name;
	d[2] = toString(locus[i]->bp);

	d[3] = locus[i]->allele1;
	d[4] = locus[i]->allele2;
	if (locus[i]->allele1 == 0) d[3] = "-";
	if (locus[i]->allele2 == 0) d[4] = "-";

	d[5] = toString(locus[i]->maf, "%.3f");
	d[6] = toString(locus[i]->mafCases, "%.3f");
	d[7] = toString(locus[i]->mafControls, "%.3f");;

	data.push_back(d);
    }

    printTable(out, data);
}

// Read file containing exclusion list of SNPs
void readSNPExclusions()
{
    string s;
    ifstream in;
    int countExclude = 0, countExtract = 0;

    // Read in file for --exclude

    if (excludefile.size() != 0) {

	in.open(excludefile.c_str());

	cout << "Reading SNP exclusion list from file " + excludefile + "\n";

	if (!in) {
	    cout << "Error: Cannot open file " + excludefile + "\n";
	    gracefulExit(1);
	}

	while (in >> s) {
	    for (int i = 0; i < nlocus; i++)
		if (locus[i]->name == s && locus[i]->id != -1) {
		    // cout << "Excluding snp " << s << "\n";
		    locus[i]->id = -1;
		    countExclude++;
		}
	}

	in.close();
    }

    // Read in file for --extract

    if (extractfile.size() != 0) {

	cout << "Reading list of SNPs to keep from file " + extractfile + "\n";

	in.open(extractfile.c_str());
	if (!in) {
	    cout << "Error: Cannot open file " + extractfile + "\n";
	    gracefulExit(1);
	}

	map<string, int> snplist;

	while (in >> s)
	    snplist[s] = 1;

	for (int i = 0; i < nlocus; i++)
	    if (snplist[locus[i]->name] != 1) {
		// cout << "Excluding snp " << locus[i]->name	 << " (not in selected SNPs)\n";
		locus[i]->id = -1;
		countExtract++;
	    }
    }

    if (countExclude > 0) cout << toString(countExclude) + " SNPs were excluded from file "
	    + excludefile + "\n";

    if (countExtract > 0) cout << countExtract << " SNPs were deleted, not present in file "
	    + extractfile + "\n";

    deleteSNPS();
    cout.flush();
}

void filtering()
{
    //if (mafFilter == -1) return;

    if (debug) cout << "Filtering SNPs based on maf = " << mafFilter << " " << " maf in cases = "
	    << mafFilterCases << " maf in controls = " << mafFilterControls << "\n";

    int a1, a2;
    int AFFa1, AFFa2, UNAFFa1, UNAFFa2;

    int one, two;
    int del = 0;

    for (int j = 0; j < locus.size(); j++) {
	a1 = 0;
	a2 = 0;
	AFFa1 = AFFa2 = 0;
	UNAFFa1 = UNAFFa2 = 0;

	for (int i = 0; i < persons.size(); i++) {

	    //one = persons[i]->one[j];
	    //two = persons[i]->two[j];
	    int geno = persons[i]->getGeno(j);

	    if (geno == G_aa) {
		a1 += 2;
	    } else if (geno == G_AA) {
		a2 += 2;
	    } else if (geno == G_aA) {
		a1++;
		a2++;
	    }

	    // Count in controls
	    if (persons[i]->phenotype == 1) {
		if (geno == G_aa) {
		    UNAFFa1 += 2;
		} else if (geno == G_AA) {
		    UNAFFa2 += 2;
		} else if (geno == G_aA) {
		    UNAFFa1++;
		    UNAFFa2++;
		}
	    }

	    // Count in cases
	    if (persons[i]->phenotype == 2) {
		if (geno == G_aa) {
		    AFFa1 += 2;
		} else if (geno == G_AA) {
		    AFFa2 += 2;
		} else if (geno == G_aA) {
		    AFFa1++;
		    AFFa2++;
		}
	    }

	}

	double maf = a1 / ((double) a1 + a2);
	double Maf = a2 / ((double) a1 + a2);

	double mafControls = UNAFFa1 / ((double) UNAFFa1 + UNAFFa2);
	double MafControls = UNAFFa2 / ((double) UNAFFa1 + UNAFFa2);

	double mafCases = AFFa1 / ((double) AFFa1 + AFFa2);
	double MafCases = AFFa2 / ((double) AFFa1 + AFFa2);

	locus[j]->maf = min(maf, Maf);
	locus[j]->mafCases = min(mafCases, MafCases);
	locus[j]->mafControls = min(mafControls, MafControls);

	if(QuantTrait == 0) locus[j]->maf = locus[j]->mafControls;

	// Mark snp for deletion, according to filtering rules
	if ( locus[j]->maf < mafFilter ||
	     locus[j]->mafCases < mafFilterCases ||
	     locus[j]->mafControls < mafFilterControls) {

	    locus[j]->id = -1;
	    del++;
	}

	/*cout << "Snp " << locus[j]->name <<  " " << locus[j]->allele1 << " "
	 << " " << " maf = " << maf << " " <<   locus[j]->allele2 << " " <<
	 " Maf = " << Maf << " " << (maf+Maf) << "\n";*/

    }

    if (del > 0) {
	cout << "Deleting " << del << " SNPs failing maf frequency test (maf < "
	     << mafFilter << ").\n";

	deleteSNPS();
    }

    cout.flush();
}

// Completely remove SNPS that are marked as deleted

void deleteSNPS()
{
    // printf("Warning!!!!!!!! Delete SNPs is not ready!\n");
    // SNPs marked with id=-1 are deleted

    int c = 0;

    for (int j = 0; j < locus.size(); j++)
	if (locus[j]->id != -1) c++;

    if (c == locus.size()) return;

    if (debug) cout << "Removing " << (locus.size() - c) << " snps.\n";

    // Remove SNPS genotype data
    for (int i = 0; i < persons.size(); i++) {

	Individual *p = persons[i];
	Individual *n = new Individual(*p);

	vector<int> temp;

	for (int j = 0; j < locus.size(); j++)
	    if (locus[j]->id != -1) temp.push_back(p->getGeno(j));

	n->resizeGeno(temp.size());

	for (int j = 0; j < temp.size(); j++)
	    n->setGeno(j, temp[j]);

	persons[i] = n;
	delete p;
    }

    // Create a new locus vector and free memory not used
    vector<Locus*> newlocus;

    for (int j = 0; j < locus.size(); j++)
	if (locus[j]->id != -1) newlocus.push_back(locus[j]);
	else delete locus[j];

    locus = newlocus;
    nlocus = locus.size();

    for (int j = 0; j < locus.size(); j++)
	locus[j]->id = j;

    if (debug) cout << "   Done (" << nlocus << " snps left)\n";
}



void joinClusterRuns()
{

    if (clusterJoin == 1) {
	cout << "\nJoining the results from cluster runs..\n";
    }

    vector<vector<Results> > L0;

    for (int i = 1; i <= clusterN ; i ++) {
	ifstream clusterFile;

	string fname = resultsFile + "." + toString(i) + "." + toString(clusterN);
	clusterFile.open(fname.c_str());

	cout << "Reading partial results from file " << fname << ".\n";

	if(! clusterFile ) {
	    cout << "Error: Cannot open cluster results file: " << fname << "\n";
	    gracefulExit(1);
	}

	readL0(clusterFile, L0);
	readL1(clusterFile);



    }

    // Compute level 0 results

    int N = L0.size();

    cout << "L0 size is " << N << "\n";
    if(L0.size() > 0) cout << "L0[0] is " << L0[0].size() << "\n";

    // TODO: there is duplicated truncation code here. It would be better if it was integrated in computeBias

    // Compute Uboot0[k] k=1..Sn

    Uboot0 = computeBias(Un, L0, 1);

    // Truncation of Uboot at 0
    if (!notruncate) for (int i = 0; i < Sn; i++)
	if (Uboot0[i] != NA && Uboot0[i] < 0) Uboot0[i] = 0;


    // Hopefully should work

    computeFinalResults();

    outputFinalResults();

}


/*
 * runBootstrap(): main function to perform double level bootstrap
 */

void runBootStrap()
{
    remove(resultsFile);

    if(clusterJoin == 1) { joinClusterRuns(); return; }

    vector<Locus*> selected; // Which snps are selected for br2 analysis

    selected = getNaiveEstimate();

    getNaiveOR(persons, selected);

    getLevel0Quantities();

    getLevel12Quantities();

    if(clusterN > 0) {
        cout << "\n\n";
        cout << "Cluster iteration " << clusterK << " finished.\n";
        cout << "Use --join when all instances are done to get the final results.\n";

        gracefulExit(0);
    }

    computeFinalResults();

    outputFinalResults();

    return;
}

/*
 * write_single_files(): write single snp files.
 */

void writeSingleSNPFiles()
{
    vector<Locus*> significant; // Which snps are selected for br2 analysis

    significant = getNaiveEstimate();

    for (int i = 0; i < significant.size(); i++) {

	cout << "Writing single SNP : " << significant[i]->name << "\n";
	vector<Locus*> temp;
	temp.push_back(significant[i]);

	plinkRunner.writeBinFiles("snp-" + toString((double) i, "%0.0f"), persons, temp, 0);

    }
}

/*
 * Delete all the temporary files created
 *
 */
void deleteTempFiles()
{
    if (debug) return;

    char prefix[10][50] = {
    "all", "ud", "ue", "un"
    };

    char ext[8][50] = {
    ".bed", ".bim", ".fam", ".model", ".assoc.logistic", ".assoc.linear", ".log", ".cov"
    };

    int threads = numCpu;

    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 8; j++) {
	    char filename[1024];
	    int k = threads - 1;

	    while (k >= 0) {
		sprintf(filename, "%stemp_%d%s%s", tmpdir.c_str(), k, prefix[i], ext[j]);
		remove(filename);
		k--;
	    }

	    sprintf(filename, "%stemp_%s%s", tmpdir.c_str(), prefix[i], ext[j]);
	    remove(filename);
	}

    // Additional files to remove
    remove(tmpdir + "plinktest.map");
    remove(tmpdir + "plinktest.ped");
    remove(tmpdir + "plinktest.log");

    remove(tmpdir + "lockfile");
}

void cleanupMemory()
{
    for (int i = 0; i < persons.size(); i++)
	delete persons[i];

    for (int i = 0; i < locus.size(); i++)
	delete locus[i];
}

void gracefulExit(int retval)
{
    deleteTempFiles();
    exit(retval);
}

void help()
{
    cout << "\nShelly Bull, Apostolos Dimitromanolakis, Laura Faye, Lei Sun\n";
    cout << "\n";
    cout << "Usage:\n";
    cout << "\n";
    cout << "   br2 <options>\n";
    cout << "\n";
    cout
	    << "br2 is a command line utility to perform bootstrap-based bias reduction in the GWA setting. The input is a .ped/.map file \n";
    cout
	    << "combination and the bias-reduced estimates of the significant SNPs in the input file is written to the output file out.br2. \n";
    
    cout << "\n";
    cout << "Options:\n";
    cout << "\n";
    cout << "Selecting input/output files:\n";
    cout << "   --file basefilename \n";
    cout << "      Selects input files basefilename.ped and  basefilename.map as input files.\n";
    cout << "\n";
    cout << "   --ped pedfilename\n";
    cout << "      Selects ped file pedfilename for input.\n";
    cout << "\n";
    cout << "   --map mapfilename\n";
    cout << "      Selects map file mapfilename for input.\n";
    cout << "\n";
    cout << "   Either --file or both --map and --ped have to be specified.\n";
    cout << "\n";
    cout << "   --out outputfilename\n";
    cout
	    << "      Writes the final values to the specified file. The default is to write to file out.br2\n";
    cout << "\n";
    cout << "Specifying parameters for the analysis:\n";
    cout << "   --alpha val\n";
    cout
	    << "      Uses alpha cutoff value val for to select significant SNPs. val can be specified \n";
    cout
	    << "      using either normal or scientific notation (for example –alpha 0.001 or –alpha 1e-6).\n";
    cout << "\n";
    cout << "   --top num\n";
    cout
	    << "      Selects for analysis at most the top num SNPs from the original input file, ordered by p-value.\n";
    cout << "\n";
    cout << "   --keep num\n";
    cout
	    << "      Reduce the number of SNPs to speedup computation. The top num significant SNPs are used          \n";
    cout << "      in the analysis. All the other are ignored.\n";
    cout << "\n";
    cout << "   --test testTYPE\n";
    cout << "            Select a specific test type for significance. Can be one of: \n";
    cout << "         minTRENDGENO: selects the minimum p-value of the \n";
    cout << "           genotypic and trend tests.\n";
    cout << "         GENO: selects the genotypic test p-value\n";
    cout << "         TREND: selects the trend test p-value\n";
    cout << "         LOGISTIC: selects the logistic test p-value \n";
    cout << "         ALLELIC: basic allelic test\n";

    cout << "\n";
    cout << "      Warning: The logistic test is much slower than all the others. \n";
    cout << "\n";
    cout << "   --B1 num\n";
    cout << "      Specify number of first level bootstraps to perform.\n";
    cout << "\n";
    cout << "   --B2 num\n";
    cout << "      Specify number of second level bootstraps to perform for SD. \n";
    cout << "\n";
    cout << "\n";
    cout << "\n";
    cout << "\n";
    cout << "System options:\n";
    cout << "   --scratch dir\n";
    cout << "      Selects a directory to store temporary files.\n";
    cout << "\n";
    cout << "   --verbose \n";
    cout
	    << "      Be verbose and print significance data and OR for every iteration of the bootstrap.\n";
    cout << "\n";
    cout << "   --cpu num\n";
    cout
	    << "      If compilied with multi-CPU support then this options forces the use of exactly num cpu's.          \n";
    cout << "      Otherwise the number of CPUs to use is auto-detected.\n";
    cout << "\n";
    cout << "   --seed num\n";
    cout << "      Selects a specific seed value for the random number generator. Runs with the \n";
    cout << "      same seed value should produce identical results.\n";
    cout << "\n";
    cout << "   --plink path\n";
    cout
	    << "      Specify the path to the plink executable, if plink is not in your PATH variable.\n";
    cout << "\n";
    cout << "\n";
    cout << "Examples:\n";
    cout << "\n";
    cout << "   br2 --file input --alpha 1e-7 --l0 100 --l1 100 --l2 100\n";
    cout << "   \n";
    cout
	    << "   br2 --ped input.ped --map input.map --alpha 1e-10 --scratch /tmp --out br2.output\n";
    cout << "\n";
    cout
	    << "   br2 --file input --cpu 5 --plink /software/plink/plink --l0 50 --l1 50 --l2 50 --maf 0.05 \n";
    cout << "      --exclude snps.txt\n";
    cout << "\n";
}


void filter_chromosome(int chr)
{
    cout << "Selecting only chromosome " << chr << " for analysis.\n";

    for (int i = 0; i < nlocus; i++)
	if (locus[i]->id != -1) if (locus[i]->chr != chr || locus[i]->bp == 0) {
	    locus[i]->id = -1;
	}

    deleteSNPS();
}

/*
 * should be deleted later
 */
void alternative_test()
{
    int N;

    vector<vector<double> > pval;
    vector<double> orig;
    vector<ModelOutput> out;

    pval.resize(locus.size() + 1);
    orig.resize(locus.size() + 1);

    isSignificant("tmp1", persons, locus, 1, out);

    for (int i = 0; i < out.size(); i++) {

	int id = out[i].snp->id;
	double p = out[i].p;

	orig[id] = p;
    }

    for (N = 0; N < Level0Bootstraps; N++) {
	BootStrap L0BS(&persons);
	L0BS.generateBootstrap_2();
	cout << N << "\n";

	isSignificant("tmp1", L0BS.BootstrapSample, locus, 1, out);

	for (int i = 0; i < out.size(); i++) {

	    int id = out[i].snp->id;
	    double p = out[i].p;

	    pval[id].push_back(p);
	}

    }

    char s[2000];
    ofstream outF("alt.txt");

    for (int i = 0; i < locus.size(); i++) {
	double av = avg(pval[i]);

	vector<double> l;
	for (int j = 0; j < pval[i].size(); j++)
	    l.push_back(log(pval[i][j]));

	double logav = avg(l);

	sprintf(s, "%16s %10g %10g %10g\n", locus[i]->name.c_str(), orig[i], av, logav);

	outF << s;
    }

}

int main(int argc, char** argv)
{
    char s[2000];
    cout << version_string << "\n";

    // cout << "-----------------\nWARNING: EXPERIMENTAL VERSION\n-------------\n";

    #ifdef _OPENMP

    numCpu = omp_get_num_procs();

    sprintf(s,"  - multi-processor enabled (OpenMP) -\n  - machine has %d processors - \n\n", numCpu);
    cout << s;

    // cout << "CPU=" << omp_get_num_procs() << " THREADS=" << omp_get_max_threads() << "\n";

    #endif

    if (sizeof(int*) == 8) cout << "  - running in 64-bit mode - \n";
    if (sizeof(int*) == 4) cout << "  - running in 32-bit mode - \n";

    cout << "\n";

    // Initialize random number generators

    int seed = time(NULL);
    srand(seed);

    RNG.Reset();
    RNG.SetSeed(seed);

    int chr = -1; // Specifies if we only do single chromosome analysis
    int singleSNP = 0; // If we want to do single SNP analysis on each significant SNP

    int makeFemale = 0;
    int randomPheno = 0;
    int runBenchmark = 0;
    int altTest = 0;

    string baseFileName = "input", ped, map;

    // Read in command line options

    options opts(argc, argv);

    if (opts.get("--file", 1)) {
	baseFileName = opts.arg(0);
	ped = baseFileName + ".ped";
	map = baseFileName + ".map";
    }

    if (opts.get("--ped", 1)) {
	if (ped.size() != 0) {
	    cout << "Error: please specify only one of --file or --ped.\n";
	    exit(1);
	}

	ped = (string) opts.arg(0);
    }

    if (opts.get("--map", 1)) {
	if (map.size() != 0) {
	    cout << "Error: please specify only one of --file or --map.\n";
	    exit(1);
	}

	map = (string) opts.arg(0);
    }

    if (opts.get("--test", 1)) {
	if (strcmp(opts.arg(0), testTypes[0]) == 0) testType = 0;
	else if (strcmp(opts.arg(0), testTypes[1]) == 0) testType = 1;
	else if (strcmp(opts.arg(0), testTypes[2]) == 0) testType = 2;
	else if (strcmp(opts.arg(0), testTypes[3]) == 0) testType = 3;
	else if (strcmp(opts.arg(0), testTypes[4]) == 0) testType = 4;
	else if (strcmp(opts.arg(0), testTypes[5]) == 0) testType = 5;
	else if (strcmp(opts.arg(0), testTypes[6]) == 0) testType = 6;
	else if (strcmp(opts.arg(0), testTypes[7]) == 0) testType = 7;
	else {
	    cout << "Error: Unknown test type \"" << opts.arg(0) << "\"\n" << "  should be one of ";
	    for(int i=0;i<7;i++) cout << testTypes[i] << " ";
	    cout << "\n";
	    exit(1);
	}

    }

    if (opts.get("--estimate", 1)) {
	if (strcmp(opts.arg(0), "LOGISTIC") == 0) estimationType = LOGISTIC_MODEL;
	else if (strcmp(opts.arg(0), "ALLELIC") == 0) estimationType = ALLELIC_MODEL;
	else if (strcmp(opts.arg(0), "DUMMY") == 0) estimationType = DUMMY_MODEL;
	else if (strcmp(opts.arg(0), "RPLUGIN") == 0) estimationType = RPLUGIN_MODEL;

	else {
	    cout << "Error: Unknown estimate type \"" << opts.arg(0) << "\"\n"
		    << "  should be one of LOGISTIC, ALLELIC or RPLUGIN" << "\n";
	    exit(1);
	}

    }

    if (opts.get("--qt", 0)) QuantTrait = 1;

    if (opts.get("--R", 1)) {
	RScriptFile = opts.arg(0);
    }

    if (opts.get("--L0", 1)) Level0Bootstraps = atoi(opts.arg(0));
    if (opts.get("--L1", 1)) Level1Bootstraps = atoi(opts.arg(0));
    if (opts.get("--L2", 1)) Level2Bootstraps = atoi(opts.arg(0));

    // Alternative way to specify iteration numbers as suggested by Lei
    if (opts.get("--B1", 1)) {
        Level0Bootstraps = atoi(opts.arg(0));
        Level2Bootstraps = atoi(opts.arg(0));
    }

    if (opts.get("--B2", 1)) Level1Bootstraps = atoi(opts.arg(0));

    if (opts.get("--alpha", 1)) alpha = atof(opts.arg(0));
    if (opts.get("--top", 1)) maxSNPs = atoi(opts.arg(0));

    if (opts.get("--maf", 1)) mafFilter = atof(opts.arg(0));

    if (opts.get("--strictmaf", 1)) {
	mafFilter = atof(opts.arg(0));
	mafFilterCases = atof(opts.arg(0));
	mafFilterControls = atof(opts.arg(0));
    }

    if (opts.get("--fast", 0)) usePlinkCode = 1;
    if (opts.get("--slow", 0)) usePlinkCode = 0;

    if (opts.get("--plink", 1)) plinkRunner.plinkLocation = opts.arg(0);

    if (opts.get("--random", 0)) randomPheno = 1;
    if (opts.get("--makefemale", 0)) makeFemale = 1;

    if (opts.get("--silent", 0)) silent = 1;
    if (opts.get("--verbose", 0)) silent = 0;
    if (opts.get("--debug", 0)) debug++;

    if (opts.get("--allbias", 0)) logAllBias = 1;

    if (opts.get("--benchmark", 1)) runBenchmark = atoi(opts.arg(0));

    if (opts.get("--out", 1)) resultsFile = opts.arg(0);

    if (opts.get("--stats", 0)) printStatsFile = 1;

    if (opts.get("--notruncate", 0)) notruncate = 1;

    if (opts.get("--cell", 1)) {
	plinkRunner.plinkOptions += " --cell " + toString(atoi(opts.arg(0)));
	par::min_geno_cell = atoi(opts.arg(0)); // for internal code from plink
    }

    if (opts.get("--help", 0)) {
	help();
	exit(0);
    }

    if (opts.get("--seed", 1)) {
	seed = atoi(opts.arg(0));
	cout << "Setting random seed to " << seed << "\n";
	RNG.SetSeed(seed);
    }

    if (opts.get("--single", 0)) singleSNP = 1;

    if (opts.get("--chr", 1)) chr = atoi(opts.arg(0));

    if (opts.get("--exclude", 1)) excludefile = opts.arg(0);

    if (opts.get("--extract", 1)) extractfile = opts.arg(0);

    if (opts.get("--keep", 1)) keeptop = atoi(opts.arg(0));

    if (opts.get("--keeprandom", 1)) keeprandom = atoi(opts.arg(0));

    if (opts.get("--mafcorrection", 0)) mafCorrection = 1;
    if (opts.get("--nomafcorrection", 0)) mafCorrection = 0;

    if (opts.get("--log", 0)) reportLogOR = 1;

    if (opts.get("--scratch", 1)) {
	tmpdir = opts.arg(0);
	// Add a final slash to scratch directory
	if (tmpdir[tmpdir.size() - 1] != '/') tmpdir += "/";
    }

    if (opts.get("--cpu", 1)) {

    #ifdef _OPENMP
	numCpu = atoi(opts.arg(0));
	cout << "Setting number of cpus to use to " << numCpu << ".\n\n";
	omp_set_num_threads(numCpu);
    #else
	cout << "Option --cpu is only useful in multi-processor version. Ignoring it now. \n";
    #endif
    }

    if (opts.get("--pheno", 1)) {
	par::alternative_phenotype_file = opts.arg(0);
    }

    if (opts.get("--missing-phenotype", 1)) {
    	par::missing_phenotype = opts.arg(0);
    }

    if (opts.get("--useplink", 0)) {
	useInternalRegression = 0;
    }

    if (opts.get("--covar", 1)) {
	haveCovariates = true;
	par::covar_filename = opts.arg(0);
	// par::covariate_command_line += "--covar " + (string)opts.arg(0);
    }

    if (opts.get("--covar-number", 1)) {
	haveCovariates = true;
	par::clist_selection_number = true;
	par::clist_selection_string = opts.arg(0);
	par::covariate_command_line += "--covar-number " + (string) opts.arg(0);
    }

    if (opts.get("--covar-name", 1)) {
	if (par::clist_selection_number == true) {
	    cout
		    << "Error: Multiple covariate selection commands. Use either --covar-number or --covar-name\n";
	    gracefulExit(1);
	}

	haveCovariates = true;
	par::clist_selection_name = true;
	par::clist_selection_string = opts.arg(0);
	par::covariate_command_line += "--covar-name " + (string) opts.arg(0);
    }

    if (opts.get("--alt", 1)) altTest = 1;

    if (opts.get("--dummy", 2)) {
	DummyDataTests::generate(atoi(opts.arg(0)), atoi(opts.arg(1)));
	dummyRun = 1;
	//testType = DUMMY;
    }

    if (opts.get("--cluster", 1)) {
        int n = sscanf(opts.arg(0), "%d/%d", &clusterK, &clusterN);
        if(n != 2) {
            cout << "Error: --cluster option accepts a string of the type k/n (e.g. 1/20)\n";
            gracefulExit(1);
        }
    }

    if (opts.get("--join",1)) {

        clusterJoin = 1;
        clusterN = atoi(opts.arg(0));
    }

    // Check and report errors for unused command line options
    opts.cleanup();


    // Test for consistency of command line options.

    if (dummyRun == 0) if (ped.size() == 0 || map.size() == 0) {
	cout << "Error: Please specify input file using "
	    "--file <basefilename> or --ped <pedfile> and --map <mapfile> \n";

	exit(1);
    }

    if (haveCovariates == true) {
	if (par::covar_filename == "") {
	    cout << "Error: No covariate file specified (please use --covar).\n";
	    gracefulExit(1);
	}

	if (QuantTrait == 1) {
	}
	else
	if (testType != LOGISTIC) {
	    cout << "Error: When using covariates, you must use logistic test (--test LOGISTIC)\n";
	    gracefulExit(1);
	}
    }

    cout << "\nCommand line options are: ";
    opts.print();
    cout << "\n\n";


    if (clusterJoin == 1 && clusterK > 0) {
	cout << "Error: Use only one of --cluster or --join.\n";
	gracefulExit(1);
    }


    if (clusterJoin == 1) {
	cout << "\nJoining the results from cluster runs..\n";
    }


    // Prepare for cluster run
    if (clusterK != 0) {
        string fname = resultsFile + "." + toString(clusterK) + "." + toString(clusterN);
        clusterFile.open(fname.c_str());

        // Try to distribute bootstraps evenly

        Level0Bootstraps = divide( clusterK, clusterN, Level0Bootstraps);
        Level1Bootstraps = divide( clusterK, clusterN, Level1Bootstraps);

        // Choose an appropriate starting seed.

        seed = seed + clusterK * 7777; // random constant
        RNG.SetSeed(seed);


        cout << "Clustering instance " << clusterK << " of " << clusterN << "\n";
        cout << "  Will run " << Level0Bootstraps << " level 0 bootstraps.\n";
        cout << "  Will run " << Level1Bootstraps << " level 1 bootstraps.\n";
        cout << "  Using seed " << seed << " for random number generation.\n";
    }

    // Make br2 user proof:
    // Add a random number in front of temporary files prefix in case users decide to run br2
    // in parallel using the same temporary directory

    sprintf(s, "%07d", rand()%1000000 );
    tmpdir += "temp.br2." + (string)s + ".";

    if(clusterN > 0) {
	tmpdir += toString(clusterK) + ".";
    }


    // Check to see if we need to run plink.

    if(usePlinkCode == 1) {
	// Check for plink and get plink version

	double plinkVersion = plinkRunner.checkPlinkVersion(tmpdir);

	if (plinkVersion < 0) {
	    cout << "Error: Cannot run plink."
		"Please install plink and check that it is in your path.\n\n";

	    exit(1);
	}

	cout << "Detected plink version " << plinkVersion << "\n";

	if (plinkVersion != 1.04 && plinkVersion != 1.05) {
	    cout << "Warning: br2 has only been tested with versions 1.04/1.05 of plink."
		" It may not work with current version.\n";
	}
    }

    cout << "\n";


    sprintf(s, "Analysis parameters:\n");
    cout << s;
    sprintf(s, "   input pedigree file = %s (--ped)\n", ped.c_str());
    cout << s;
    sprintf(s, "   input map file = %s (--map)\n", map.c_str());
    cout << s;
    sprintf(s, "   output file = %s (--out)\n", resultsFile.c_str());
    cout << s;

    if(QuantTrait == 0)
	sprintf(s, "   test type = %s\n", testTypes[testType]);
    else
	sprintf(s, "   test type = %s\n", "LINEAR");

    cout << s;

    sprintf(s, "   estimation = ");
    cout << s;

    if (QuantTrait == 0 && estimationType == LOGISTIC_MODEL) cout << "LOGISTIC\n";
    if (QuantTrait == 0 && estimationType == GENOTYPIC_MODEL) cout << "GENOTYPIC\n";
    if (QuantTrait == 0 && estimationType == ALLELIC_MODEL) cout << "ALLELIC\n";
    if (QuantTrait == 0 && estimationType == DUMMY_MODEL) cout << "DUMMY\n";

    if (estimationType == RPLUGIN_MODEL)
	cout << "R PLUGIN " << RScriptFile << "\n";
    else
	if (QuantTrait == 1) cout << "LINEAR\n";

    sprintf(s, "   phenotype = %s\n", (QuantTrait == 1) ? "quantitative" : "binary (case/controls)");
    cout << s;

    sprintf(s, "   alpha = %g (--alpha)\n", alpha);
    cout << s;

    if (maxSNPs != -1) {
	sprintf(s, "   max SNPs = %d (--top)\n", maxSNPs);
	cout << s;
    }

    sprintf(s, "   bootstraps = %d + %d x %d\n", Level0Bootstraps, Level1Bootstraps,
	    Level2Bootstraps);
    cout << s;

    if (haveCovariates == true) sprintf(s, "   covariates = yes (file: %s, index: %s)\n",
	    par::clist_selection_string.c_str(), par::clist_selection_string.c_str());

    cout << "\n";

    if(mafCorrection) cout << "Using MAF correction procedure.\n";

    cout << "Reading input files.\n";

    cout.flush();

    if (dummyRun == 0) plinkRunner.readTextFiles(ped, map);

    // Read alternative phenotype file
    if (par::alternative_phenotype_file != "")
	plinkRunner.readAltPhenotype(par::alternative_phenotype_file);

    // Read in covariates
    if (haveCovariates) plinkRunner.readCovariateFile(par::covar_filename);

    // Make all individuals females if requested
    if (makeFemale == 1) for (int i = 0; i < npersons; i++)
	persons[i]->sexcode = "1";

    // Generate Random phenotype if requested
    if (randomPheno == 1) for (int i = 0; i < npersons; i++)
	persons[i]->phenotype = RNG.GetRand(1, 2);

    int controls = 0, cases = 0;

    // Set persons with phenotype != 1 and phenotype != 2 to missing
    if(QuantTrait == 0)
      for(int i=0; i<npersons; i++)
	  if(persons[i]->phenotype != 1 && persons[i]->phenotype != 2)
	      persons[i]->phenotype = NA;

    // Print some statistics about the phenotype

    if (QuantTrait == 0) {

	for (int i = 0; i < npersons; i++) {
	    persons[i]->phenotype == 1 && controls++;
	    persons[i]->phenotype == 2 && cases++;
	}

	cout << "Binary trait:\n";
	cout << "    " << cases << " cases.\n";
	cout << "    " << controls << " controls.\n";
	cout << "    " << (npersons - controls - cases) << " missing.\n";

    } else {

	double min = 1e40, max = -1e40;
	vector <double> f;

	for (int i = 0; i < npersons; i++)
	    if (persons[i]->phenotype != NA) {
		if (persons[i]->phenotype < min) min = persons[i]->phenotype;
		if (persons[i]->phenotype > max) max = persons[i]->phenotype;
		f.push_back( persons[i]->phenotype );
	    }

	cout << "Quantitative trait:\n";
	cout << "    " << " minimum " + toString(min, "%.3g") << "\n";
	cout << "    " << " maximum " + toString(max, "%.3g") << "\n";
	cout << "    " << " average " + toString(avg(f), "%.3g") << "\n";
	cout << "    " << "      sd " + toString(sd(f), "%.3g") << "\n";
    }

    cout << "\n";

    plinkRunner.deleteMissingPersons();

    cout.flush();

    // Perform filtering

    if (chr != -1) filter_chromosome(chr);

    filtering();
    readSNPExclusions();

    cout << "SNPs remaining after filtering: " << locus.size() << "\n";

    cout.flush();

    if (printStatsFile == 1) printStats();

    if (singleSNP == 1) {
	writeSingleSNPFiles();
	gracefulExit(0);
    }

    if (altTest == 1) {
	alternative_test();
	exit(0);
    }

    if (estimationType == GENOTYPIC_MODEL) Genotypic::runBootStrap();
    else runBootStrap();

    // Write snp rank information
    if (debug && clusterN == 0) {
	ofstream out2;
	out2.open("summary.txt");

	for (int i = 0; i < locus.size(); i++) {
	    vector<double> &v = locus[i]->ranks;

	    if (0) {
		ofstream out1;

		out1.open(("snp-" + locus[i]->name).c_str());

		for (int j = 0; j < v.size(); j++)
		    out1 << v[j] << " ";
		out1 << "\n";
		out1.close();

	    }

	    int sig = 0;
	    for (int j = 0; j < v.size(); j++)
		if (v[j] <= Sn) sig++;

	    out2 << locus[i]->name << " " << locus[i]->p << " ";
	    out2 << locus[i]->rank << " " << min(v) << " " << max(v) << " " << avg(v) << " " << sd(
		    v) << " " << sig << "\n";
	}
    }

    // Print Warning message if we found large OR
    if (flagBadOR > 0) {
	cout << "Warning: Some SNPs were found to have OR > "
	    << (int) (exp(WarnLogORLimit) + 0.5) << ".\n"
	    "  This is probably due to sparse/missing data.\n"
	    "  Please check the log file for the names of the offending SNPs (search for Warning:).\n";
    }

    cleanupMemory();
    deleteTempFiles();
    return 0;
}

