/*
 * br2.h:  Global Data Structure Definitions
 *
 *  Created on: Jan 29, 2009
 *  Author: Apostolos Dimitromanolakis
 */

#ifndef BR2_H_
#define BR2_H_

#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <algorithm>
#include <map>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "random_gen.h"
#include "utils.h"
#include "genotypic.h"

using namespace std;

namespace br2options {
    extern int flagBadOR, Level0Bootstraps, Level1Bootstraps;
    extern int Level2Bootstraps, silent, maxSNPs, keeptop;
    extern int QuantTrait, debug;

    extern double WarnLogORLimit, alpha;

    extern string resultsFile;
    extern bool haveCovariates;
};

namespace par {
    extern string covariate_header;
    extern string covariate_command_line;
    extern string missing_phenotype;
    extern vector<string> covariate_names;
};

extern int nlocus;
extern CRandom RNG;

using namespace br2options;

// Genotype coding: 0 = aa (a1 hom) , 1 = MM (missing) , 2 = aA (a1/a2) , 3 = AA (a2/a2)
enum GENOTYPE
{
    G_aa = 0,    G_AA = 3,    G_aA = 2,    G_MM = 1
};

/**
 *  Class Locus contains all the information for a SNP in input file
 *
 */

class Locus
{
public:
    // Data as specified in plink .map file

    int id; // Serial Number of this Locus, position in Locus table.
    int chr; // Chromosome number
    double pos; // cM map positions
    int bp; // base-pair position

    string name; // ID for this locus

    char allele1, allele2; // Minor and major allele

    int rank; // Rank of this snp for the naive estimate
    vector<double> ranks; // Ranks of this SNP for every bootstrap iteration
    int numSig; // number of times this snp was significant

    double p; // Naive P value

    double maf, mafCases, mafControls; // minor allele frequency

    Locus()
    {
	chr = 0;
	name = "";
	allele1 = 0;
	allele2 = 0;
	pos = 0;
	bp = 0;
	rank = 0;
	numSig = 0;
    }

    ~Locus()
    {
	// cout << "Deleting snp " << name << "\n";
    }

    bool operator<(const Locus & p2) const
    {
	return (chr < p2.chr || (chr == p2.chr && bp < p2.bp));
    }

    bool operator==(const Locus & p2) const
    {
	return (name == p2.name);
    }

};


template<class A>
class svector : public vector<A> {
public:
    ~svector() {
	cout << "Destructing an svector of size " << this->size() << "\n";
    }
};

/*
 *  Data for a single individual as read from the .ped file
 */

class Individual
{
public:
    int id; // number of this individual in input file

    string fid; // family id
    string iid; // individual id

    // Parental codes
    string pat; //  Paternal ID
    string mat; //  Maternal ID

    string sexcode; // Sex code 1=male 2=female other=unknown
    double phenotype; // 0=missing 1=unaffected 2=affected OR quantitative

    vector<double> covariates; // Covariates for this individual

    Individual()
    {
    }

    ~Individual()
    {
    }

    Individual(const Individual &a)
    {
	// cout << "Copy constructor for " << a.id << " " << a.geno.size() << " " << geno.size() << "\n";
	id = a.id;
	fid = a.fid;
	iid = a.iid;
	pat = a.pat;
	mat = a.mat;
	sexcode = a.sexcode;
	phenotype = a.phenotype;
	covariates = a.covariates;
    }

    // Genotype data

#if 0
    vector<unsigned char> one;

    void resizeGeno(int n) {one.resize(n);}

    void setGeno(int pos, int g) { one[pos] = g; }

    int getGeno(int pos) {
	return one[pos];
    }

#endif


    // Select which data structure to use for genotype data. The default should be the fastest

    #if 0
    vector<bool> one;

    void resizeGeno(int n) {one.resize(2*n+2);}

    void setGeno(int pos, int g) {
	one[pos<<1] = g&1;
	one[(pos<<1)+1] = g>>1;
    }

    int getGeno(int pos) {
	return one[pos<<1] + (one[(pos<<1)+1]<<1);
    }

    // 32bit /2 = 16 -> 4 bits
    #else

    typedef unsigned long bit_type;

    vector<bit_type> geno;

    bool getOne(int n)
    {
	return getGeno(n) >> 1;
    }

    bool getTwo(int n)
    {
	return getGeno(n) & 1;
    }

    void resizeGeno(int n)
    {
	//printf("size = %d %d\n", (2+n/(4*sizeof(bit_type))), sizeof(bit_type));
	geno.resize(2 + n / (4 * sizeof(bit_type)));

	//geno = new bit_type[(2+n/(4*sizeof(bit_type)))];
    }

    int getGeno(int pos)
    {
	if (sizeof(bit_type) == 4) {
	    bit_type p = pos >> 4;
	    bit_type m = (pos & 0xf) << 1;
	    return (geno[p] >> m) & 3;
	} else if (sizeof(bit_type) == 8) {
	    bit_type p = pos >> 5;
	    bit_type m = (pos & 0x1f) << 1;
	    return (geno[p] >> m) & 3;
	} else {
	    cout << "Unsupported computer architecture: sizeof(unsigned int) = "
		    << sizeof(bit_type) << "\n";
	}
    }

    void setGeno(int pos, int g)
    {
	if (sizeof(bit_type) == 4) {
	    bit_type p = pos >> 4;
	    bit_type m = (pos & 0xf) << 1;
	    geno[p] = (geno[p] & (~((bit_type) 3 << m))) | ((bit_type) g << m);
	} else {
	    bit_type p = pos >> 5;
	    bit_type m = (pos & 0x1f) << 1;
	    geno[p] = (geno[p] & (~((bit_type) 3 << m))) | ((bit_type) g << m);
	}
    }

    #endif

};


inline
bool Individual_sortByPos(Individual *p1, Individual *p2)
{
    return p1 < p2;
    return (p1->id < p2->id);
}



/**
 *
 *  Class BootStrap:
 *    Used to select the bootstrap sample and compute the bias.
 *    Also now is aware if we are running QTL.
 *
 */

class BootStrap
{
public:
    // The selected individuals for the bootstrap samples
    vector<Individual*> BootstrapSample;

    // Individuals not selected for current bootstrap
    vector<Individual*> NotSelected;

private:

    int ncases, ncontrols;
    int *cases, *controls;

    // Individuals from which to select bootstrap sample

    vector<Individual*> *persons_p;
    int npersons;

public:
    BootStrap(vector<Individual*> *mypersons)
    {
	persons_p = mypersons;
	npersons = persons_p->size();

	vector<Individual*> & persons = *persons_p;

	ncases = 0;
	ncontrols = 0;

	if (QuantTrait == 1) return;

	cases = new int[npersons];
	controls = new int[npersons];

	// Count number of controls & cases
	for (int i = 0; i < npersons; i++) {
	    if (persons[i]->phenotype == 1) controls[ncontrols++] = i;
	    if (persons[i]->phenotype == 2) cases[ncases++] = i;
	}

	// cout << "Bootstrapping: Controls=" << ncontrols << " Cases=" << ncases	<< "\n";

	BootstrapSample.reserve(npersons);
	NotSelected.reserve(npersons);
    }

    ~BootStrap()
    {
	if (QuantTrait == 1) return;

	delete[] cases;
	delete[] controls;
    }

    /**
     * Generate a bootstrap sample for QT, sampling from all individuals
     */
    void generateBootstrapQTL()
    {
	vector<Individual*> & persons = *persons_p;

	// Indicator of selected individuals
	vector<int> selected;
	selected.resize(npersons);

	for (int i = 0; i < npersons; i++) selected[i] = 0;

	BootstrapSample.clear();
	NotSelected.clear();

	BootstrapSample.reserve(npersons);
	NotSelected.reserve(npersons);

	// The improved procedure speeds up br2 by 10-20% due to optimized memory access patterns.

	for (int i = 0; i < npersons; i++) {
	    int p = RNG.GetRand(0, npersons - 1);
	    selected[p] ++;
	}

	for (int i = 0; i < npersons; i++)
	    if (selected[i] == 0) NotSelected.push_back(persons[i]);

	for (int i = 0; i < npersons; i++)
	    for(int j = 0; j < selected[i]; j++)
		BootstrapSample.push_back(persons[i]);


	if (debug) cout << "QTL Bootstrap : Sample=" << BootstrapSample.size() << " NotSel="
		<< NotSelected.size() << " " << (NotSelected.size() / (double) npersons) << "\n";
    }

    /*
     * Alternative sampling for bootstrap (shouldn't be using, only for test purposes)
     */

    void generateBootstrap_2()
    {
	cout << "Warning: Using debugging version of bootstrap code\n";

	vector<Individual*> & persons = *persons_p;

	// Indicator of selected individuals
	vector<int> selected;
	selected.resize(npersons);
	for (int i = 0; i < npersons; i++) selected[i] = 0;

	BootstrapSample.clear();
	NotSelected.clear();

	// first ncontrol entries of BootstrapSample are random sample with replacement from controls
	// next ncase entries of BootstrapSample are random sample with replacement from patients

	//Select with replacement ncases individuals from cases
	for (int i = 0; i < npersons; i++) {
	    int p = RNG.GetRand(0, 10000);
	    if(p > 5000) {
		BootstrapSample.push_back(persons[i]);
		BootstrapSample.push_back(persons[i]);
		selected[i] = 1;
	    }

	    else { NotSelected.push_back(persons[i]); NotSelected.push_back(persons[i]);}

	    /*p = NCBIRandom.GetRand(0, 1);
	    if(p==1) NotSelected.push_back(persons[i]);
	    if(p==1) NotSelected.push_back(persons[i]);*/
	}

	/*//Select with replacement ncontrols individuals from controls
	for (int i = 0; i < ncontrols; i++) {
	    int p = NCBIRandom.GetRand(0, 1);
	    if(p==1) BootstrapSample.push_back(persons[controls[i]]);

	    p = NCBIRandom.GetRand(0, 1);
	    if(p==1) NotSelected.push_back(persons[controls[i]]);
	}*/

	/*cout << "Sample: " << BootstrapSample.size() << " " << NotSelected.size() << "\n";
	for (int i = 0; i < BootstrapSample.size(); i++) cout << BootstrapSample[i]->iid << " ";
	cout << "\n";
	for (int i = 0; i < NotSelected.size(); i++) cout << NotSelected[i]->iid << " ";*/

    }

    /**
     *  Generate one random bootstrap sample, sampling independendtly from cases and controls
     */

    void generateBootstrap()
    {
	if (QuantTrait == 1) {
	    generateBootstrapQTL();
	    return;
	}

	vector<Individual*> & persons = *persons_p;

	vector<int> selected; // selected individuals

	selected.resize(npersons);
	for (int i = 0; i < npersons; i++) selected[i] = 0;

	BootstrapSample.clear();
	NotSelected.clear();

	// first ncontrol entries of BootstrapSample are random sample with replacement from controls
	// next ncase entries of BootstrapSample are random sample with replacement from patients

	//Select with replacement ncases individuals from cases
	for (int i = 0; i < ncases; i++) {
	    int p = RNG.GetRand(0, ncases - 1);
	    selected[cases[p]]++;
	}

	//Select with replacement ncontrols individuals from controls
	for (int i = 0; i < ncontrols; i++) {
	    int p = RNG.GetRand(0, ncontrols - 1);
	    selected[controls[p]]++;
	}

	// The improved procedure results in a sorted sample
	// and speeds up br2 by 10-20% due to optimized memory access patterns.

	for (int i = 0; i < npersons; i++)
	    if ( selected[i] == 0 ) NotSelected.push_back(persons[i]);

	for (int i = 0; i < npersons; i++)
	    for(int j = 0; j < selected[i]; j++)
		BootstrapSample.push_back(persons[i]);
    }

};

class GenotypicOutput;

/*
 *  Perform necessary adjustments for naive OR
 */
inline
void adjust(Locus *snp, double &OR, char &a1)
{
    // if odds ratio < 1 then odds=1/odds, switch allele
    if (OR < 1) {
	OR = 1 / OR;

	if (a1 == snp->allele1) a1 = snp->allele2;
	else if (a1 == snp->allele2) a1 = snp->allele1;
    }
}

/*
 *  Perform necessary adjustments for Ud,Ue OR
 */
inline
void adjust(double &UdOR, char &UdA1, double &UeOR, char &UeA1)
{
    // If one of Ud/Ue is NA set both to NA and do nothing
    if (UeOR == NA || UdOR == NA) {
	UeOR = NA;
	UdOR = NA;
	return;
    }

    if (UeOR < 0 || UdOR < 0) {
	cout << "Negative odds ratio found! \n";
	gracefulExit(1);
    }

    // if A1 allele is same for both D and E steps keep OR same

    // if A1 allele is different , flip OR for the Estep
    if (UdA1 != UeA1) {
	UeOR = 1 / UeOR;
	UeA1 = UdA1;
    }

    // if Ud and Ue odds ratio go in different directios set Ue to 1
    if (((UdOR < 1) && (UeOR > 1)) || ((UdOR > 1) && (UeOR < 1))) {
	UeOR = 1;
    }

    // if Ud<1 then change direction for both Ue and Ud
    if (UdOR < 1) {
	UdOR = 1 / UdOR;
	UeOR = 1 / UeOR;
    }

}

class GenotypicResults
{
public:
    Locus *snp; // Reference pointer to locus[] matrix
    double p; // Significance p-value

    char a1; // Minor allele

    char UdA1, UeA1; // Minor allele for the D and E step
    double UdOR[2], UeOR[2]; // Ud and Ue log odds ratios

    GenotypicResults(Locus *s, double _p)
    {
	snp = s;
	p = _p;
    }

    GenotypicResults()
    {
	UdOR[0] = UdOR[1] = NA;
	snp = NULL;
	p = 10;
    }

    // Necessary adjustment for the naive OR estimate
    void adjustNaive()
    {
	if (UdOR[0] == NA || UdOR[1] == NA) {
	    UdOR[0] = NA;
	    UdOR[1] = NA;
	    return;
	}

	::adjust(snp, UdOR[0], UdA1);
	::adjust(snp, UdOR[1], UdA1);

	UdOR[0] = log(UdOR[0]);
	UdOR[1] = log(UdOR[1]);

	if (UdOR[0] > WarnLogORLimit || UdOR[1] > WarnLogORLimit) {
	    cout << "Warning: Log Odds Ratio > " << WarnLogORLimit << " found for snp: "
		    << snp->name << " ADDOR=" << UdOR[0] << " DOMDEVOR=" << UdOR[1] << "\n";
	    flagBadOR = 1;
	}
    }

    void adjust()
    {
	/*
	 *  Perform necessary adjustments for Ud,Ue steps. Flip OR if necessary and compute log OR
	 */
	::adjust(UdOR[0], UdA1, UeOR[0], UeA1);
	::adjust(UdOR[1], UdA1, UeOR[1], UeA1);

	// Compute log Odds ratios
	UdOR[0] = log(UdOR[0]);
	UdOR[1] = log(UdOR[1]);
	UeOR[0] = log(UeOR[0]);
	UeOR[1] = log(UeOR[1]);

	if (UdOR[0] > WarnLogORLimit || UdOR[1] > WarnLogORLimit || UeOR[0] > WarnLogORLimit
		|| UeOR[1] > WarnLogORLimit) {
	    cout << "Warning: Log Odds Ratio > " << WarnLogORLimit << " found for snp: "
		    << snp->name << " UdOR=" << UdOR[0] << "," << UdOR[1] << " UeOR=" << UdOR[0]
		    << "," << UeOR[1] << "\n";

	    flagBadOR = 1;
	}
    }

    /*void getOR(vector<GenotypicOutput> &results, const string &head, vector<Individual*> &persons, vector<Locus*> &locus)
     {
     plinkRunner.writeBinFiles(head, persons, locus, 1);
     plinkRunner.Genotypic(head, tmpdir + "temp_un", results, locus);
     }*/

};

inline bool GenotypicResults_sortByP(const GenotypicResults &lhs, const GenotypicResults &rhs)
{
    return lhs.p < rhs.p;
}

inline bool GenotypicResults_sortByCHR(const GenotypicResults &lhs, const GenotypicResults &rhs)
{
    return (lhs.snp->chr < rhs.snp->chr || (lhs.snp->chr == rhs.snp->chr && lhs.snp->bp
	    < rhs.snp->bp));
}

inline bool GenotypicResults_sortByLocus(const GenotypicResults &lhs, const GenotypicResults &rhs)
{
    return (lhs.snp < rhs.snp);
}

/*
 * class Results:
 *    holds resulting data about a particular bootstrap iteration and a particular snp.
 *
 */

class Results
{
public:
    Locus *snp; // Reference pointer to locus[] matrix
    double p; // Significance p-value

    double odds; // Used to Store odds ratio returned by plink
    char a1; // Minor allele

    char UdA1, UeA1; // Minor allele for the D and E step
    double UdOR, UeOR; // Ud and Ue log odds ratios

    Results()
    {
	odds = NA;
	UdOR = NA;
	UeOR = NA;
    }
    Results(Locus *s, double _p)
    {
	snp = s;
	p = _p;
    }

    void adjustNaive()
    {
	if (odds == NA) return;

	// if odds ratio < 1 then odds=1/odds, switch allele
	if (odds < 1) {
	    odds = 1 / odds;

	    if (a1 == snp->allele1) a1 = snp->allele2;
	    else if (a1 == snp->allele2) a1 = snp->allele1;
	}

	odds = log(odds);

	if (odds > WarnLogORLimit) if(QuantTrait == 0) if(flagBadOR < 200) {
	    cout << "Warning: Log Odds Ratio > " << WarnLogORLimit << " found for snp: "
		    << snp->name << " OR=" << odds << "\n";
	    flagBadOR ++;
	}

    }

    void adjust()
    {
	/*
	 *  Perform necessary adjustments. Flip OR if necessary and compute log OR
	 */

	// If one of Ud/Ue is NA set both to NA and do nothing
	if (UeOR == NA || UdOR == NA) {
	    UeOR = NA;
	    UdOR = NA;
	    return;
	}

	if (UeOR < 0 || UdOR < 0) {
	    cout << "Negative odds ratio found! \n";
	    gracefulExit(1);
	}

	//cout << "Adjust 1: " << (UdOR) << "," << UdA1 << " " << (UeOR) << "," << UeA1 << "\n";

	// if A1 allele is same for both D and E steps keep OR same

	// if A1 allele is different , flip OR for the Estep
	if (UdA1 != UeA1) {
	    UeOR = 1 / UeOR;
	    UeA1 = UdA1;
	}

	// if Ud and Ue odds ratio go in different directios set Ue to 1
	if (((UdOR < 1) && (UeOR > 1)) || ((UdOR > 1) && (UeOR < 1))) {
	    UeOR = 1;
	}

	// if Ud<1 then change direction for both Ue and Ud
	if (UdOR < 1) {
	    UdOR = 1 / UdOR;
	    UeOR = 1 / UeOR;
	}

	//cout << "Adjust 2: " << (UdOR) << "," << UdA1 << " " << (UeOR) << "," << UeA1 << "\n";

	// Compute log Odds ratios
	UdOR = log(UdOR);
	UeOR = log(UeOR);

	if (UdOR > WarnLogORLimit || UeOR > WarnLogORLimit)
	    if(QuantTrait == 0) if(flagBadOR < 200) {
		cout << "Warning: Log Odds Ratio > " << WarnLogORLimit << " found for snp: "
		    << snp->name << " UdOR=" << UdOR << " UeOR=" << UeOR << "\n";

		flagBadOR ++;
	    }
    }
};

inline bool Results_sortByP(const Results &lhs, const Results &rhs)
{
    return lhs.p < rhs.p;
}

inline bool Results_sortByCHR(const Results &lhs, const Results &rhs)
{
    return (lhs.snp->chr < rhs.snp->chr || (lhs.snp->chr == rhs.snp->chr && lhs.snp->bp
	    < rhs.snp->bp));
}

inline bool Results_sortByLocus(const Results &lhs, const Results &rhs)
{
    return (lhs.snp < rhs.snp);
}

inline
bool Locus_sortByCHR(Locus*p1, Locus *p2)
{
    return (p1->chr < p2->chr || (p1->chr == p2->chr && p1->bp < p2->bp));
}


/*
 * class LogisticOutput:
 * 		Stores information returned by plink --logistic
 *
 */

class LogisticOutput
{
public:
    Locus *snp; // Reference pointer to locus[] matrix

    char a1; // Minor allele
    double odds; // Odds ratio

    /* Model Output Data */
    double p; // p value

    LogisticOutput(Locus *_snp, char _a1, double _odds, double _p)
    {
	odds = _odds;
	a1 = _a1;
	snp = _snp;
	p = _p;
    }

    LogisticOutput()
    {
	snp = NULL;
	p = odds = NA;
    }

    void print()
    {
	cout << snp->name << " " << a1 << " " << p << " " << odds << " " << "\n";
    }
};

inline bool LogisticOutput_sortByLocus(const LogisticOutput &lhs, const LogisticOutput &rhs)
{
    return (lhs.snp < rhs.snp);
}

/*
 * class LogisticResults:
 * 		Stores information returned by plink --logistic
 *
 */

class GenotypicOutput
{
public:
    Locus *snp; // Reference pointer to locus[] matrix

    char a1; // Minor allele
    double ADD; // Additive OR
    double DOMDEV; // Dominance deviation OR
    double p; // p value

    GenotypicOutput(Locus *_snp, char _a1, double _ADD, double _DOM, double _p)
    {
	a1 = _a1;
	snp = _snp;
	ADD = _ADD;
	DOMDEV = _DOM;
	p = _p;
    }

    GenotypicOutput()
    {
	clear();
    }

    void clear()
    {
	snp = NULL;
	ADD = DOMDEV = p = NA;
    }

    void print()
    {
	cout << snp->name << " " << a1 << " P=" << p << " ADD=" << ADD << " DOM=" << DOMDEV << "\n";
    }
};

inline bool GenotypicOutput_sortByLocus(const GenotypicOutput &lhs, const GenotypicOutput &rhs)
{
    return (lhs.snp < rhs.snp);
}

/*
 * class ModelResults:
 * 		Stores information returned by plink --model
 */

class ModelOutput
{
public:
    Locus *snp; // Reference number to locus[] matrix

    int significant; // 1 if this snp is significant

    /* Model Output Data */
    char a1;
    double pTrend, pGeno;

    double p;

    ModelOutput()
    {
	snp = NULL;
	a1 = 0;
    }
    
};

inline bool operator<(const ModelOutput &lhs, const ModelOutput &rhs)
{
    return lhs.p < rhs.p;
}

inline bool ModelResults_sortbyCHR(const ModelOutput &lhs, const ModelOutput &rhs)
{
    return (lhs.snp->chr < rhs.snp->chr || (lhs.snp->chr == rhs.snp->chr && lhs.snp->bp
	    < rhs.snp->bp));
}

inline bool Locus_sortbyP(const Locus *a, const Locus *b)
{
    return a->p < b-> p;
}



/*
 *  Extern forward function/variable declerations
 *
 */

vector<Locus*> getNaiveEstimate();
void getNaiveOR(vector<Individual*> & persons, vector<Locus*> &selectedSNPS);
void deleteSNPS();

int isSignificant(const string &infile, vector<Individual*> &Sample, vector<Locus*> &SNPs,
	int newfids, vector<ModelOutput> &out);

extern int nlocus;
extern vector<Locus*> locus;

// Persons is a list of all individuals along with all the genotype data
extern int npersons;
extern vector<Individual*> persons;


#endif /* BR2_H_ */
